#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_backup_and_restore.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import shutil
import socket
import sys
import tarfile
import time
import re
import base64
from logging.handlers import RotatingFileHandler
from os import listdir
from os.path import isfile, join

# Third-party libraries
import requests

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import (
    trackme_getloglevel,
    run_splunk_search,
    trackme_reqinfo,
    trackme_create_report,
    trackme_create_alert,
    trackme_delete_report,
    trackme_create_kvcollection,
    trackme_delete_kvcollection,
    trackme_create_kvtransform,
    trackme_delete_kvtransform,
    trackme_create_macro,
    trackme_delete_macro,
)

# import TrackMe get data libs
from trackme_libs_get_data import (
    get_full_kv_collection,
)

# import Splunk libs
import splunklib.client as client

#
# Splunk Cloud certification notes: these functions create and manage archive files for the application in the following directory $SPLUNK/etc/apps/trackme/backup
# There are no options to perform any kind of file manipulations out of this application directory
#


class TrackMeHandlerBackupAndRestore_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerBackupAndRestore_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_backup_and_restore(self, request_info, **kwargs):
        response = {
            "resource_group_name": "backup_and_restore",
            "resource_group_desc": "These endpoints provide backup and restore facilities for the Kvstore collections created and managed in TrackMe, this includes the full scope of active and enabled tenants",
        }

        return {"payload": response, "status": 200}

    # List backup archive files known from the KV, add any local file that the KV wouldn't do about
    def get_backup(self, request_info, **kwargs):
        """
        A simple function to safetly retrieve all records from a KVstore collection with pagination
        """

        def get_full_collection_records(collection):
            collection_records = []
            collection_records_dict = {}
            collection_records_keys = set()

            end = False
            skip_tracker = 0
            while not end:
                process_collection_records = collection.data.query(skip=skip_tracker)
                if process_collection_records:
                    for item in process_collection_records:
                        collection_records.append(item)
                        collection_records_dict[item.get("_key")] = (
                            item  # Add the entire item to the dictionary
                        )
                        collection_records_keys.add(item.get("_key"))
                    skip_tracker += 1000
                else:
                    end = True

            return collection_records, collection_records_dict, collection_records_keys

        describe = False

        # init
        mode = "full"

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False

            # mode summary
            try:
                mode = resp_dict["mode"]
                # valid mode: full | summary (defaults to full)
                if mode not in ("full", "summary"):
                    return {
                        "payload": f'Invalid mode="{mode}", valid modes are: full | summary',
                        "status": 400,
                    }
            except Exception as e:
                mode = "full"

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        if describe:
            response = {
                "describe": "This endpoint lists all the backup files available on the search head, files are stored in the TrackMe backup directory, it requires a GET call with the following arguments:",
                "resource_desc": "Get the list of backups known to TrackMe",
                "resource_spl_example": '| trackme mode=get url="/services/trackme/v2/backup_and_restore/backup"',
                "options": [
                    {
                        "mode": "(string) OPTIONAL: The output mode, valid values are full | summary, defaults to full",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        else:
            # Get splunkd port
            splunkd_port = request_info.server_rest_port

            # set loglevel
            loglevel = trackme_getloglevel(
                request_info.system_authtoken, request_info.server_rest_port
            )
            log.setLevel(logging.getLevelName(loglevel))

            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # get local server name
            server_name = socket.gethostname()

            # check backup dir existence
            if not os.path.isdir(backuproot):
                logging.info(
                    f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                )
                return {
                    "payload": {
                        "response": f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                    },
                    "status": 200,
                }

            else:
                # store files in list
                backup_files = [
                    join(backuproot, f)
                    for f in listdir(backuproot)
                    if isfile(join(backuproot, f)) and f.endswith(".tgz")
                ]

                if not backup_files:
                    logging.info(
                        f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                    )
                    return {
                        "payload": {
                            "response": f'TrackMe backup process, there are no backup archives available in directory="{backuproot}", instance="{server_name}"'
                        },
                        "status": 200,
                    }

                else:
                    # Enter the list, verify for that for each archive file we have a corresponding record in the audit collection
                    # if there is no record, then we create a replacement record
                    for backupfile in backup_files:
                        # get the file mtime
                        try:
                            backup_file_mtime = round(os.path.getmtime(backupfile))
                        except Exception as e:
                            backup_file_mtime = None
                            logging.error(
                                f'failure to retrieve the tmime of the tar file archive="{backupfile}" with exception:"{str(e)}"'
                            )

                        # get the file size
                        try:
                            tar_filesize = os.path.getsize(backupfile)
                        except Exception as e:
                            tar_filesize = None
                            logging.error(
                                f'archive="{backupfile}", failure to retrieve the size of the tar file archive="{str(e)}"'
                            )

                        # Store a record in a backup audit collection

                        # Create a message
                        status_message = "discovered"

                        # record / key
                        record = None
                        key = None

                        # Define the KV query
                        query_string = {
                            "$and": [
                                {
                                    "backup_archive": backupfile,
                                    "server_name": server_name,
                                }
                            ]
                        }

                        # backup audit collection
                        collection_name_backup_archives_info = (
                            "kv_trackme_backup_archives_info"
                        )
                        service_backup_archives_info = client.connect(
                            owner="nobody",
                            app="trackme",
                            port=splunkd_port,
                            token=request_info.session_key,
                            timeout=300,
                        )
                        collection_backup_archives_info = (
                            service_backup_archives_info.kvstore[
                                collection_name_backup_archives_info
                            ]
                        )

                        try:
                            kvrecord = collection_backup_archives_info.data.query(
                                query=json.dumps(query_string)
                            )[0]
                            key = kvrecord.get("_key")

                        except Exception as e:
                            key = None

                        # If a record cannot be found, this backup file is not know to TrackMe currently, add a new record
                        if key is None:
                            # attempt to extract and get the backup details

                            # Attempt extraction
                            backup_file_is_valid = False

                            # From TrackMe 2.1.5, we generate a metadata file named as <tarfile>.full.meta and <tarfile>.light.meta
                            # Check if the metadata file exists
                            backup_file_has_metadata = False

                            # check for both metadata files
                            if os.path.isfile(
                                f"{backupfile}.full.meta"
                            ) and os.path.isfile(f"{backupfile}.light.meta"):
                                backup_file_has_metadata = True

                                try:
                                    with open(
                                        f"{backupfile}.full.meta", "r"
                                    ) as read_content:
                                        full_metadata = json.load(read_content)
                                except Exception as e:
                                    full_metadata = None
                                    backup_file_has_metadata = False
                                    logging.error(
                                        f'failed to load the full metadata file="{backupfile}.full.meta" with exception="{str(e)}"'
                                    )

                                try:
                                    with open(
                                        f"{backupfile}.light.meta", "r"
                                    ) as read_content:
                                        light_metadata = json.load(read_content)
                                except Exception as e:
                                    light_metadata = None
                                    backup_file_has_metadata = False
                                    logging.error(
                                        f'failed to load the light metadata file="{backupfile}.light.meta" with exception="{str(e)}"'
                                    )

                            # Set backup dir
                            backupdir = os.path.join(
                                backuproot, os.path.splitext(backupfile)[0]
                            )

                            try:
                                tf = tarfile.open(backupfile)
                                tf.extractall(backupdir)
                                backup_file_is_valid = True
                            except Exception as e:
                                response = {
                                    "response": f'The archive name {backupfile} could not be extracted, restore cannot be processed, this backup file will not be considered, exception="{str(e)}"',
                                }
                                logging.error(json.dumps(response, indent=2))

                            #
                            # backup file generated with TrackMe 2.1.5 and later
                            #

                            if backup_file_is_valid and backup_file_has_metadata:

                                logging.info(
                                    f'Discovered backup file="{backupfile}" with metadata compatibility mode.'
                                )

                                if full_metadata and light_metadata:

                                    try:
                                        # Insert the record
                                        collection_backup_archives_info.data.insert(
                                            json.dumps(
                                                {
                                                    "mtime": full_metadata.get("mtime"),
                                                    "htime": full_metadata.get("htime"),
                                                    "server_name": full_metadata.get(
                                                        "server_name"
                                                    ),
                                                    "status": json.dumps(
                                                        light_metadata, indent=4
                                                    ),
                                                    "change_type": "backup archive was missing from the info collection and added by automatic discovery (metadata compatibility mode)",
                                                    "backup_archive": full_metadata.get(
                                                        "backup_archive"
                                                    ),
                                                    "size": full_metadata.get("size"),
                                                    "archive_details": json.dumps(
                                                        full_metadata.get(
                                                            "archive_details"
                                                        ),
                                                        indent=4,
                                                    ),
                                                }
                                            )
                                        )

                                    except Exception as e:
                                        logging.error(
                                            f'failed to insert a new KVstore record with exception="{str(e)}"'
                                        )
                                        return {
                                            "payload": {
                                                "response": f'failed to insert a new KVstore record with exception="{str(e)}"'
                                            },
                                            "status": 500,
                                        }

                            #
                            # backup file generated prior to TrackMe 2.1.5
                            #

                            elif backup_file_is_valid:

                                logging.info(
                                    f'Discovered backup file="{backupfile}" with legacy compatibility mode.'
                                )

                                # store the list of json files in a list
                                collections_json_files = [
                                    f
                                    for f in listdir(backupdir)
                                    if isfile(join(backupdir, f))
                                ]

                                # store the list of available collections in the archive
                                collections_available = []

                                # create a dictionnary
                                collections_restore_dict = {}
                                for json_file in collections_json_files:
                                    # strip the extension
                                    collection_name = os.path.splitext(json_file)[0]

                                    # append to the list of available collections for restore
                                    collections_available.append(collection_name)

                                    # get the file size
                                    json_file_size = os.path.getsize(
                                        os.path.join(backupdir, json_file)
                                    )

                                    # get the file mtime
                                    json_file_mtime = round(
                                        os.path.getmtime(
                                            os.path.join(backupdir, json_file)
                                        )
                                    )

                                    # try getting the number of records
                                    try:
                                        with open(
                                            os.path.join(backupdir, json_file), "r"
                                        ) as read_content:
                                            json_file_records = len(
                                                json.load(read_content)
                                            )
                                    except Exception as e:
                                        json_file_records = None

                                    # add to the dict
                                    collections_restore_dict[collection_name] = {
                                        "file": json_file,
                                        "size": json_file_size,
                                        "mtime": json_file_mtime,
                                        "records": json_file_records,
                                    }

                                # remove backup dir
                                try:
                                    shutil.rmtree(backupdir)
                                except OSError as e:
                                    logging.error(
                                        f'failed to purge the extraction temporary directory="{backupdir}" with exception="{str(e)}"'
                                    )

                                try:
                                    # Insert the record
                                    collection_backup_archives_info.data.insert(
                                        json.dumps(
                                            {
                                                "mtime": str(backup_file_mtime),
                                                "htime": str(
                                                    time.strftime(
                                                        "%c",
                                                        time.localtime(
                                                            backup_file_mtime
                                                        ),
                                                    )
                                                ),
                                                "server_name": str(server_name),
                                                "status": str(status_message),
                                                "change_type": "backup archive was missing from the info collection and added by automatic discovery (legacy archive compatibility mode)",
                                                "backup_archive": str(backupfile),
                                                "size": str(tar_filesize),
                                                "archive_details": json.dumps(
                                                    collections_restore_dict,
                                                    indent=4,
                                                ),
                                            }
                                        )
                                    )

                                except Exception as e:
                                    logging.error(
                                        f'failed to insert a new KVstore record with exception="{str(e)}"'
                                    )
                                    return {
                                        "payload": {
                                            "response": f'failed to insert a new KVstore record with exception="{str(e)}"'
                                        },
                                        "status": 500,
                                    }

                    # Render

                    records = collection_backup_archives_info.data.query()
                    currently_known_archives = []
                    for record in records:
                        currently_known_archives.append(record.get("backup_archive"))

                    logging.info(
                        f'TrackMe get backup files finished successfully, archive_fields="{currently_known_archives}"'
                    )

                    (
                        collection_records,
                        collection_records_dict,
                        collection_records_keys,
                    ) = get_full_collection_records(collection_backup_archives_info)

                    if mode == "full":
                        return {
                            "payload": collection_records,
                            "status": 200,
                        }

                    elif mode == "summary":

                        backup_count = len(collection_records)
                        backup_files = []

                        for record in collection_records:
                            backup_files.append(record.get("backup_archive"))

                        return {
                            "payload": {
                                "backup_count": backup_count,
                                "backup_files": backup_files,
                            },
                            "status": 200,
                        }

    # Take a backup
    def post_backup(self, request_info, **kwargs):

        describe = False

        """
        A simple function to safetly retrieve all records from a KVstore collection with pagination
        """

        def get_full_collection_records(collection):
            collection_records = []
            collection_records_dict = {}
            collection_records_keys = set()

            end = False
            skip_tracker = 0
            while not end:
                process_collection_records = collection.data.query(skip=skip_tracker)
                if process_collection_records:
                    for item in process_collection_records:
                        collection_records.append(item)
                        collection_records_dict[item.get("_key")] = (
                            item  # Add the entire item to the dictionary
                        )
                        collection_records_keys.add(item.get("_key"))
                    skip_tracker += 1000
                else:
                    end = True

            return collection_records, collection_records_dict, collection_records_keys

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False

            # comment
            try:
                comment = resp_dict["comment"]
                if len(comment) == 0:
                    comment = None
            except Exception as e:
                comment = None
            if not comment:
                comment = f"Backup initiated by {request_info.user}, date: {time.strftime('%c')}"

        else:
            # body is not required in this endpoint, if not submitted do not describe the usage
            describe = False

        if describe:
            response = {
                "describe": "This endpoint performs a backup of all TrackMe collections in a compressed tarball, this file is stored in the TrackMe backup directory, it requires a POST call the following arguments.",
                "resource_desc": "Start TrackMe backup process",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/backup_and_restore/backup"',
                "options": [
                    {
                        "comment": "OPTIONAL: comment to be added to the backup archive",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        else:
            # Get splunkd port
            splunkd_port = request_info.server_rest_port

            # Get service
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )

            # header
            header = {
                "Authorization": "Splunk %s" % request_info.session_key,
                "Content-Type": "application/json",
            }

            # set loglevel
            loglevel = trackme_getloglevel(
                request_info.system_authtoken, request_info.server_rest_port
            )
            log.setLevel(logging.getLevelName(loglevel))

            # Get trackmeconf
            trackme_conf = trackme_reqinfo(
                request_info.system_authtoken, request_info.server_rest_uri
            )
            trackme_version = trackme_conf.get("trackme_version", "unknown")

            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # Set timestr
            timestr = time.strftime("trackme-backup-%Y%m%d-%H%M%S")

            # Set backup dir
            backupdir = os.path.join(backuproot, str(timestr))

            # get local server name
            server_name = socket.gethostname()

            # Create the backup dir if does not exist
            if not os.path.isdir(backuproot):
                os.mkdir(backuproot)
            if not os.path.isdir(backupdir):
                os.mkdir(backupdir)

            # Set up service
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.session_key,
                timeout=300,
            )

            # header
            header = {
                "Authorization": "Splunk %s" % request_info.session_key,
                "Content-Type": "application/json",
            }

            # archive details dictionnary
            archive_details = {}

            # add trackme_version to archive_details
            archive_details["trackme_version"] = trackme_version

            ############################################
            # Task 1: Connect to Virtual Tenants KVstore
            ############################################

            # Register the object summary in the vtenant collection
            collection_vtenants_name = "kv_trackme_virtual_tenants"
            collection_vtenants = service.kvstore[collection_vtenants_name]

            # Get records
            vtenants_records, vtenants_collection_keys, vtenants_collection_dict = (
                get_full_kv_collection(collection_vtenants, collection_vtenants_name)
            )

            #################################################################################
            # Task 2: For each Virtual Tenant, retrieve and backup the Virtual Tenant account
            #################################################################################

            # Iterate over the Virtual Tenants
            for vtenant_record in vtenants_records:

                # get the tenant_id
                tenant_id = vtenant_record.get("tenant_id")

                # get the status
                tenant_status = vtenant_record.get("tenant_status")

                # only consider enabled tenants
                if tenant_status == "enabled":

                    # url
                    url = f"{request_info.server_rest_uri}/services/trackme/v2/configuration/admin/get_vtenant_account"

                    try:
                        response = requests.post(
                            url,
                            headers=header,
                            data=json.dumps({"tenant_id": tenant_id}),
                            verify=False,
                            timeout=300,
                        )

                        response.raise_for_status()
                        response_json = response.json()
                        virtual_tenant_account = response_json.get("vtenant_account")

                    except Exception as e:
                        error_msg = f'failed to retrieve the list of TrackMe Knowledge Objects to be handled with exception="{str(e)}"'
                        return {
                            "payload": {
                                "response": error_msg,
                            },
                            "status": 500,
                        }

                    # save the file to our backup directory as tenant_<tenant_id>_vtenant_account.json
                    target = os.path.join(
                        backupdir, f"tenant_{tenant_id}_vtenant_account.json"
                    )

                    try:
                        with open(target, "w") as f:
                            f.write(json.dumps(virtual_tenant_account, indent=1))
                    except Exception as e:
                        logging.error(
                            f'Exception encountered while trying to backup the Virtual Tenant account for tenant_id="{tenant_id}", exception="{str(e)}"'
                        )
                        return {
                            "payload": {
                                "response": f'Exception encountered while trying to backup the Virtual Tenant account for tenant_id="{tenant_id}", exception="{str(e)}"',
                            },
                            "status": 500,
                        }

                    # add to the archive details
                    archive_details[f"tenant_{tenant_id}_vtenant_account"] = {
                        "file": f"tenant_{tenant_id}_vtenant_account.json",
                        "size": os.path.getsize(target),
                        "mtime": round(os.path.getmtime(target)),
                    }

            ##########################################
            # Task 3: Knowledge Objects identification
            ##########################################

            # total counters for reporting purposes
            total_count_tenants = 0
            total_count_kos = 0
            total_count_kvstore_collections_definitions = 0
            total_count_transforms_definitions = 0
            total_count_reports = 0
            total_count_alerts = 0
            total_count_macros = 0

            # Iterate over the Virtual Tenants
            for vtenant_record in vtenants_records:

                # get the status
                tenant_status = vtenant_record.get("tenant_status")

                # only consider enabled tenants
                if tenant_status == "enabled":

                    tenant_id = vtenant_record.get("tenant_id")
                    total_count_tenants += 1

                    # tenant counters for reporting purposes
                    tenant_count_kos = 0
                    tenant_count_kvstore_collections_definitions = 0
                    tenant_count_transforms_definitions = 0
                    tenant_count_reports = 0
                    tenant_count_alerts = 0
                    tenant_count_macros = 0

                    # a dictionary to store the list of knowledge objects
                    trackme_knowledge_objects = {}

                    # url
                    url = f"{request_info.server_rest_uri}/services/trackme/v2/configuration/get_tenant_knowledge_objects"

                    try:
                        response = requests.post(
                            url,
                            headers=header,
                            data=json.dumps({"tenant_id": tenant_id}),
                            verify=False,
                            timeout=300,
                        )

                        response.raise_for_status()
                        response_json = response.json()

                    except Exception as e:
                        error_msg = f'failed to retrieve the list of TrackMe Knowledge Objects to be handled with exception="{str(e)}"'
                        return {
                            "payload": {
                                "response": error_msg,
                            },
                            "status": 500,
                        }

                    # parse
                    for item in response_json:

                        # increment the counter
                        total_count_kos += 1
                        tenant_count_kos += 1

                        # get the type
                        type = item.get("type")

                        # get the title
                        title = item.get("title")

                        # get the properties
                        properties = item.get("properties", {})

                        # object dict
                        object_dict = {
                            "type": type,
                            "title": title,
                            "properties": properties,
                        }

                        # increment the counters
                        if type == "savedsearches":
                            total_count_reports += 1
                            tenant_count_reports += 1

                            # get the search definition
                            object_dict["definition"] = item.get("definition")

                        elif type == "alerts":
                            total_count_alerts += 1
                            tenant_count_alerts += 1

                            # get the search definition
                            object_dict["definition"] = item.get("definition")

                            # get the alert_properties
                            object_dict["alert_properties"] = item.get(
                                "alert_properties"
                            )

                        elif type == "macros":
                            total_count_macros += 1
                            tenant_count_macros += 1

                            # get the search definition
                            object_dict["definition"] = item.get("definition")

                        elif type == "lookup_definitions":
                            total_count_transforms_definitions += 1
                            tenant_count_transforms_definitions += 1

                            # get collection and fields_list
                            object_dict["collection"] = item.get("collection")
                            object_dict["fields_list"] = item.get("fields_list")

                        elif type == "kvstore_collections":
                            total_count_kvstore_collections_definitions += 1
                            tenant_count_kvstore_collections_definitions += 1

                        # add to the dict
                        trackme_knowledge_objects[title] = object_dict

                # save the file to our backup directory as knowledge_objects.json
                target = os.path.join(
                    backupdir, f"tenant_{tenant_id}_knowledge_objects.json"
                )

                try:
                    with open(target, "w") as f:
                        f.write(json.dumps(trackme_knowledge_objects, indent=1))
                except Exception as e:
                    logging.error(
                        f'Exception encountered while trying to backup the knowledge objects for tenant_id="{tenant_id}", exception="{str(e)}"'
                    )
                    return {
                        "payload": {
                            "response": f'Exception encountered while trying to backup the knowledge objects for tenant_id="{tenant_id}", exception="{str(e)}"',
                        },
                        "status": 500,
                    }

                # add to the archive details
                archive_details[f"tenant_{tenant_id}_knowledge_objects"] = {
                    "file": f"tenant_{tenant_id}_knowledge_objects.json",
                    "size": os.path.getsize(target),
                    "mtime": round(os.path.getmtime(target)),
                    "count_knowledge_objects": tenant_count_kos,
                    "count_kvstore_collections": tenant_count_kvstore_collections_definitions,
                    "count_transforms": tenant_count_transforms_definitions,
                    "count_reports": tenant_count_reports,
                    "count_alerts": tenant_count_alerts,
                    "count_macros": tenant_count_macros,
                }

            logging.info(
                f'successfully backed up the knowledge objects, total="{total_count_kos}", kvstore_collections="{total_count_kvstore_collections_definitions}", transforms="{total_count_transforms_definitions}", reports="{total_count_reports}", alerts="{total_count_alerts}", macros="{total_count_macros}"'
            )

            ####################################
            # Task 4: KVstore collections backup
            ####################################

            # to count the number of non empty collections backed up
            counter_performed = 0

            # to count the number of empty collections
            counter_empty = 0

            # retrieve the lost of collections to be backed up

            # Spawn a new search
            # Get lagging statistics from live data
            kwargs_search = {
                "app": "trackme",
                "earliest_time": "-5m",
                "latest_time": "now",
                "output_mode": "json",
                "count": 0,
            }
            searchquery = '| rest splunk_server=local "/servicesNS/nobody/trackme/storage/collections/config" | where \'eai:acl.app\'="trackme" AND disabled=0 | table title | stats values(title) as collections | tojson | fields _raw'

            # Get the results and display them using the JSONResultsReader
            try:
                # spawn the search and get the results
                reader = run_splunk_search(
                    service,
                    searchquery,
                    kwargs_search,
                    24,
                    5,
                )

                for item in reader:
                    if isinstance(item, dict):
                        collections_json = json.loads(item.get("_raw"))
                        logging.info(
                            f'successfully retrieved the list of collections to be backed up, collections_json="{collections_json}"'
                        )

            except Exception as e:
                collections_json = None
                logging.error(
                    f'failed to retrieve the list of collections to be handled with exception="{str(e)}"'
                )

            if collections_json:
                # Loop through the collections
                for collection_name in collections_json["collections"]:
                    logging.debug(f'Handling collection="{collection_name}"')

                    try:
                        collection = service.kvstore[collection_name]

                        # get records
                        (
                            collection_records,
                            collection_records_dict,
                            collection_records_keys,
                        ) = get_full_collection_records(collection)

                        # if empty, increment the counter
                        if len(collection_records) == 0:
                            counter_empty += 1

                        # Backup the collection
                        target = os.path.join(backupdir, f"{collection_name}.json")

                        try:
                            with open(target, "w") as f:
                                f.write(json.dumps(collection_records, indent=1))
                            counter_performed += 1

                        except Exception as e:
                            logging.error(
                                f'Exception encountered while trying to backup collection="{collection_name}", exception="{str(e)}"'
                            )
                            return {
                                "payload": {
                                    "response": f'Exception encountered while trying to backup collection="{collection_name}", exception="{str(e)}"',
                                },
                                "status": 500,
                            }

                    except Exception as e:
                        logging.error(
                            f'general exception encountered while trying to access to the KVstore, exception="{str(e)}"'
                        )
                        return {
                            "payload": f'general exception encountered while trying to access to the KVstore, exception="{str(e)}"',
                            "status": 500,
                        }

                # create a tgz
                tar_name = f"{backupdir}.tgz"

                try:
                    with tarfile.open(tar_name, mode="w:gz") as archive:
                        archive.add(backupdir, arcname="")

                except Exception as e:
                    logging.error(
                        f'Exception encountered while trying to generate the tar compressed archive="{tar_name}", exception="{str(e)}"'
                    )
                    return {
                        "payload": {
                            "response": f'Exception encountered while trying to generate the tar compressed archive="{tar_name}", exception="{str(e)}"',
                        },
                        "status": 500,
                    }

                # Get the archive file size
                try:
                    tar_filesize = os.path.getsize(tar_name)
                except Exception as e:
                    tar_filesize = None
                    logging.error(
                        f'tenant_id="{tar_name}", failure to retrieve the size of the tar file archive="{tar_name}" with exception:"{str(e)}"'
                    )

                # store the list of json files in a list
                collections_json_files = [
                    f
                    for f in listdir(backupdir)
                    if isfile(join(backupdir, f)) and f.startswith("kv_")
                ]

                # store the list of available collections in the archive
                collections_available = []

                # handle the dictionnary
                for json_file in collections_json_files:

                    # if the file starts by kv_
                    if json_file.startswith("kv_"):

                        if os.path.isfile(os.path.join(backupdir, json_file)):

                            # strip the extension
                            collection_name = os.path.splitext(json_file)[0]

                            # append to the list of available collections for restore
                            collections_available.append(collection_name)

                            # get the file size
                            json_file_size = os.path.getsize(
                                os.path.join(backupdir, json_file)
                            )

                            # get the file mtime
                            json_file_mtime = round(
                                os.path.getmtime(os.path.join(backupdir, json_file))
                            )

                            # try getting the number of records
                            try:
                                with open(
                                    os.path.join(backupdir, json_file), "r"
                                ) as read_content:
                                    json_file_records = len(json.load(read_content))
                            except Exception as e:
                                json_file_records = None

                            # add to the dict
                            archive_details[collection_name] = {
                                "file": json_file,
                                "size": json_file_size,
                                "mtime": json_file_mtime,
                                "records": json_file_records,
                            }

                # remove backup dir
                try:
                    shutil.rmtree(backupdir)
                except OSError as e:
                    logging.error(
                        f'failed to purge the extraction temporary directory="{backupdir}" with exception="{str(e)}"'
                    )

                # Store a record in a backup audit collection

                # backup audit collection
                collection_name_backup_archives_info = "kv_trackme_backup_archives_info"
                service_backup_archives_info = client.connect(
                    owner="nobody",
                    app="trackme",
                    port=splunkd_port,
                    token=request_info.session_key,
                    timeout=300,
                )
                collection_backup_archives_info = service_backup_archives_info.kvstore[
                    collection_name_backup_archives_info
                ]

                # Get time
                current_time = int(round(time.time()))

                # Create a message
                status_message = {
                    "trackme_version": str(trackme_version),
                    "backup_archive": str(tar_name),
                    "server_name": str(server_name),
                    "comment": str(comment),
                    "kvstore_collections_not_empty": str(counter_performed),
                    "kvstore_collections_empty": str(counter_empty),
                    "knowledge_objects_summary": {
                        "tenants_enabled_count": total_count_tenants,
                        "total": total_count_kos,
                        "kvstore_collections": total_count_kvstore_collections_definitions,
                        "transforms": total_count_transforms_definitions,
                        "reports": total_count_reports,
                        "alerts": total_count_alerts,
                        "macros": total_count_macros,
                    },
                }

                response = {
                    "mtime": str(current_time),
                    "htime": str(time.strftime("%c", time.localtime(current_time))),
                    "server_name": str(server_name),
                    "comment": str(comment),
                    "status": status_message,
                    "change_type": "backup archive created",
                    "backup_archive": str(os.path.join(backupdir, tar_name)),
                    "size": tar_filesize,
                    "archive_details": archive_details,
                }

                kvrecord = {
                    "mtime": str(current_time),
                    "htime": str(time.strftime("%c", time.localtime(current_time))),
                    "server_name": str(server_name),
                    "comment": str(comment),
                    "status": json.dumps(status_message, indent=2),
                    "change_type": "backup archive created",
                    "backup_archive": str(os.path.join(backupdir, tar_name)),
                    "size": tar_filesize,
                    "archive_details": json.dumps(archive_details, indent=2),
                }

                try:
                    # Insert the record
                    collection_backup_archives_info.data.insert(json.dumps(kvrecord))

                except Exception as e:
                    logging.error(
                        f'Exception encountered while attempting to insert a record in the KVstore collection="{collection_name_backup_archives_info}", exception="{str(e)}"'
                    )
                    return {
                        "payload": {
                            "response": f'Exception encountered while attempting to insert a record in the KVstore collection="{collection_name_backup_archives_info}", exception="{str(e)}"'
                        },
                        "status": 500,
                    }

                # In backup dir, store the status_message in a file named as the backup archive file with .full.metadata extension
                try:
                    with open(
                        f"{os.path.join(backupdir, tar_name)}.full.meta", "w"
                    ) as f:
                        f.write(json.dumps(response, indent=2))
                except Exception as e:
                    logging.error(
                        f'Exception encountered while trying to store the metadata file for the backup archive="{tar_name}", exception="{str(e)}"'
                    )
                    return {
                        "payload": {
                            "response": f'Exception encountered while trying to store the metadata file for the backup archive="{tar_name}", exception="{str(e)}"'
                        },
                        "status": 500,
                    }

                # In backup dir, store the response in a file named as the backup archive file with .light.metadata extension
                try:
                    with open(
                        f"{os.path.join(backupdir, tar_name)}.light.meta", "w"
                    ) as f:
                        f.write(json.dumps(status_message, indent=2))
                except Exception as e:
                    logging.error(
                        f'Exception encountered while trying to store the metadata file for the backup archive="{tar_name}", exception="{str(e)}"'
                    )
                    return {
                        "payload": {
                            "response": f'Exception encountered while trying to store the metadata file for the backup archive="{tar_name}", exception="{str(e)}"'
                        },
                        "status": 500,
                    }

                # Finally render a status message
                logging.info(
                    f'TrackMe backup process is terminated, status="{json.dumps(status_message, indent=4)}"'
                )
                return {"payload": {"response": response}, "status": 200}

            else:

                error_msg = "TrackMe backup process has failed, failed to retrieve the list of KVstore collections to be backed up!"
                logging.error(error_msg)
                return {
                    "payload": {
                        "response": error_msg,
                    },
                    "status": 500,
                }

    # Purge older backup archives based on a retention
    def delete_backup(self, request_info, **kwargs):

        describe = False
        retention_days = 7  # default to 7 days of retention if not specified

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # get local server name
        server_name = socket.gethostname()

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False

            # Get the mandatory backup_archive name
            try:
                retention_days = resp_dict["retention_days"]

                # convert to an integer
                try:
                    retention_days = int(retention_days)

                except Exception as e:
                    logging.error(
                        f'Invalid retention_days="{retention_days}", expecting an integer'
                    )
                    return {
                        "payload": {
                            "response": f'Invalid retention_days="{retention_days}", expecting an integer, exception="{str(e)}"',
                        },
                        "status": 500,
                    }

            except Exception as e:
                # default to 7 days
                retention_days = 7

        else:
            # body is not required in this endpoint
            describe = False

        if describe:
            response = {
                "describe": "This endpoint performs a purge of backup archive files older than x days, it requires a DELETE call with the following arguments:",
                "resource_desc": "Purge older TrackMe backup archives",
                "resource_spl_example": "| trackme url=/services/trackme/v2/backup_and_restore/backup mode=delete body=\"{'retention_days': '7'}\"",
                "options": [
                    {
                        "retention_days": "(integer) OPTIONAL: the maximal retention for backup archive files in days, if not specified defaults to 7 days",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        else:
            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # For reporting purposes
            purgedlist = []

            # retention_days = 7

            time_in_secs = time.time() - (retention_days * 24 * 60 * 60)
            for root, dirs, files in os.walk(backuproot, topdown=False):
                for file_ in files:
                    full_path = os.path.join(root, file_)
                    stat = os.stat(full_path)

                    if stat.st_mtime <= time_in_secs:
                        if os.path.isdir(full_path):
                            try:
                                os.rmdir(full_path)
                                purgedlist.append(full_path)

                            except Exception as e:
                                logging.error(
                                    f'Failed to delete the backup archive="{full_path}" with exception="{str(e)}"'
                                )
                                return {
                                    "payload": {
                                        "response": f'Failed to delete the backup archive="{full_path}" with exception="{str(e)}"',
                                    },
                                    "status": 500,
                                }

                        else:
                            if os.path.exists(full_path):
                                # delete
                                os.remove(full_path)
                                # add to list
                                purgedlist.append(full_path)

                                # Purge the record from the KVstore

                                # record / key
                                record = None
                                key = None

                                # Define the KV query
                                query_string = {
                                    "backup_archive": full_path,
                                }

                                # backup audit collection
                                collection_name_backup_archives_info = (
                                    "kv_trackme_backup_archives_info"
                                )
                                service_backup_archives_info = client.connect(
                                    owner="nobody",
                                    app="trackme",
                                    port=splunkd_port,
                                    token=request_info.session_key,
                                    timeout=300,
                                )
                                collection_backup_archives_info = (
                                    service_backup_archives_info.kvstore[
                                        collection_name_backup_archives_info
                                    ]
                                )

                                try:
                                    kvrecord = (
                                        collection_backup_archives_info.data.query(
                                            query=json.dumps(query_string)
                                        )[0]
                                    )
                                    key = kvrecord.get("_key")

                                except Exception as e:
                                    key = None

                                # If a record cannot be found, this backup file is not know to TrackMe currently, add a new record
                                if key is not None:
                                    try:
                                        # Delete the record
                                        collection_backup_archives_info.data.delete(
                                            json.dumps({"_key": key})
                                        )

                                    except Exception as e:
                                        logging.error(
                                            f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"'
                                        )
                                        return {
                                            "payload": {
                                                "response": f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"',
                                            },
                                            "status": 500,
                                        }

            if not len(purgedlist) > 2:
                response = {
                    "status": f"There were no backup files older than {retention_days} days to be purged"
                }

                logging.info(json.dumps(response, indent=4))
                return {"payload": response, "status": 200}

            else:
                # For each backup archive, if a record is found in the collection info, remove the record

                # backup audit collection
                collection_name_backup_archives_info = "kv_trackme_backup_archives_info"
                service_backup_archives_info = client.connect(
                    owner="nobody",
                    app="trackme",
                    port=splunkd_port,
                    token=request_info.session_key,
                    timeout=300,
                )
                collection_backup_archives_info = service_backup_archives_info.kvstore[
                    collection_name_backup_archives_info
                ]

                for backupfile in purgedlist:
                    key = None
                    record = None

                    # Define the KV query
                    query_string = {
                        "$and": [
                            {
                                "backup_archive": backupfile,
                                "server_name": server_name,
                            }
                        ]
                    }

                    try:
                        kvrecord = collection_backup_archives_info.data.query(
                            query=json.dumps(query_string)
                        )[0]
                        key = kvrecord.get("_key")

                    except Exception as e:
                        key = None

                    # If a record is found, it shall be purged
                    if key is not None:
                        # Remove the record
                        try:
                            collection_backup_archives_info.data.delete(
                                json.dumps({"_key": key})
                            )

                        except Exception as e:
                            logging.error(
                                f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"'
                            )
                            return {
                                "payload": {
                                    "response": f'Failed to purge the KVstore record in collection="{collection_name_backup_archives_info}" with exception="{str(e)}"',
                                },
                                "status": 500,
                            }

                response = {
                    "status": f'The following archives files were purged due to days retention="{retention_days}"',
                    "backup_files": purgedlist,
                }

                logging.info(json.dumps(response, indent=4))
                return {"payload": response, "status": 200}

    # Restore collections from a backup archive
    # For Splunk Cloud certification purposes, the archive must be located in the backup directory of the application

    def post_restore(self, request_info, **kwargs):

        def clean_backup_dir(backupdir):
            """
            A simple function to remove the backup dir
            """
            try:
                shutil.rmtree(backupdir)
                return True
            except OSError as e:
                logging.error(
                    f'failed to purge the extraction temporary directory="{backupdir}" with exception="{str(e)}"'
                )

        def restore_kvstore_records(
            service,
            collection_name,
            collections_restore_dict,
            backupdir,
            kvstore_collections_global_records_to_be_restored,
            kvstore_collections_global_records_restored,
            kvstore_collections_restored_warning,
            restore_results_dict,
            kvstore_collections_clean_empty,
        ):
            """
            Restore KVstore records from a backup JSON file, with an option to clean existing records.

            Args:
                service: Splunk service object for KVstore access.
                collection_name: The name of the KVstore collection to restore.
                collections_restore_dict: Dictionary containing metadata of collections.
                backupdir: Path to the backup directory.
                kvstore_collections_global_records_to_be_restored: Global counter for total records to be restored.
                kvstore_collections_global_records_restored: Global counter for successfully restored records.
                kvstore_collections_restored_warning: List of collections with restoration warnings.
                restore_results_dict: Dictionary to store restoration results.
                kvstore_collections_clean_empty: Boolean indicating whether to clean the collection if it was empty in the backup.

            Returns:
                A tuple containing updated global counters, as well as summary dict describing the operation:
                - kvstore_collections_global_records_to_be_restored
                - kvstore_collections_global_records_restored
                - kvstore_collection_restore_summary_dict
            """

            # Get collection metadata
            source_json_file = collections_restore_dict[collection_name]["file"]
            source_json_file_size = collections_restore_dict[collection_name]["size"]
            source_json_file_records = collections_restore_dict[collection_name][
                "records"
            ]

            # init dict summary
            kvstore_collection_restore_summary_dict = {
                "collection_name": collection_name,
                "source_json_file": source_json_file,
                "source_json_file_size": source_json_file_size,
                "source_json_file_records": source_json_file_records,
            }

            # Log start of the process
            logging.info(
                f'TrackMe restore kvstore records process starting, processing restore of collection_name="{collection_name}", '
                f'source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", '
                f'source_json_file_records="{source_json_file_records}"'
            )

            # Connect to the KVstore collection
            try:
                collection = service.kvstore[collection_name]
            except Exception as e:
                error_nsg = f'TrackMe restore kvstore records process, failed to connect to KVstore collection="{collection_name}" with exception="{str(e)}"'
                logging.error(error_nsg)
                kvstore_collections_restored_warning.append(collection_name)
                errors_list.append(error_nsg)
                return (
                    kvstore_collections_global_records_to_be_restored,
                    kvstore_collections_global_records_restored,
                    kvstore_collection_restore_summary_dict,
                )

            # Handle kvstore_collections_clean_empty
            if kvstore_collections_clean_empty and source_json_file_records == 0:
                try:
                    collection.data.delete()  # Delete all records in the collection
                    info_nsg = f'TrackMe restore kvstore records process, collection "{collection_name}" was empty in the backup. Existing records have been cleared.'
                    logging.info(info_nsg)
                    tasks_list.append(info_nsg)
                except Exception as e:
                    error_msg = f'TrackMe restore kvstore records process, failed to clear records in collection="{collection_name}" with exception="{str(e)}"'
                    logging.error(error_msg)
                    errors_list.append(error_msg)
                    kvstore_collections_restored_warning.append(collection_name)
                    return (
                        kvstore_collections_global_records_to_be_restored,
                        kvstore_collections_global_records_restored,
                        kvstore_collection_restore_summary_dict,
                    )

            # Load the JSON data
            counter = 0
            try:
                with open(os.path.join(backupdir, source_json_file), "r") as f:
                    data = json.load(f)

                # Update global counters
                kvstore_collections_global_records_to_be_restored += len(data)

                # Process data in chunks
                chunks = [data[i : i + 500] for i in range(0, len(data), 500)]
                for chunk in chunks:
                    try:
                        collection.data.batch_save(*chunk)
                        kvstore_collections_global_records_restored += len(chunk)
                        counter += len(chunk)
                    except Exception as e:
                        error_msg = (
                            f'TrackMe restore kvstore records process, failed to restore records in collection="{collection_name}" with exception="{str(e)}"'
                            f'collection_name="{collection_name}", '
                            f'source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", '
                        )
                        logging.error(error_msg)
                        errors_list.append(error_msg)
                        kvstore_collections_restored_warning.append(collection_name)

                # Add results to restore_results_dict
                restore_results_dict[collection_name] = {
                    "source_json_file": source_json_file,
                    "source_json_file_size": source_json_file_size,
                    "source_json_file_records": source_json_file_records,
                    "restored_records": counter,
                }

                # Log success or mismatch in record counts
                if int(source_json_file_records) != int(counter):
                    info_msg = (
                        f"TrackMe restore kvstore records process finished but could not verify the number of restored records "
                        f'to equal the number of records in the source file, collection_name="{collection_name}", '
                        f'source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", '
                        f'source_json_file_records="{source_json_file_records}", restored_records="{counter}"'
                    )
                    logging.error(info_msg)
                    errors_list.append(info_msg)
                else:

                    info_msg = (
                        f'TrackMe restore kvstore records process finished successfully, collection_name="{collection_name}", '
                        f'source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", '
                        f'source_json_file_records="{source_json_file_records}", restored_records="{counter}"'
                    )
                    logging.info(info_msg)
                    tasks_list.append(info_msg)

            except Exception as e:
                restore_results_dict[collection_name] = {
                    "source_json_file": source_json_file,
                    "source_json_file_size": source_json_file_size,
                    "source_json_file_records": source_json_file_records,
                    "restored_records": 0,
                    "exception": f'TrackMe restore kvstore records process, failed to open JSON file="{os.path.join(backupdir, source_json_file)}" for reading with exception="{str(e)}"',
                }
                kvstore_collections_restored_warning.append(collection_name)
                error_msg = (
                    f'TrackMe restore kvstore records process, filed to open JSON file="{os.path.join(backupdir, source_json_file)}" for reading with exception="{str(e)}"'
                    f'collection_name="{collection_name}", '
                    f'source_json_file="{source_json_file}", source_json_file_size="{source_json_file_size}", '
                    f'source_json_file_records="{source_json_file_records}", restored_records="{counter}"'
                )
                logging.error(error_msg)
                errors_list.append(error_msg)

            # add counters to the summary dict
            kvstore_collection_restore_summary_dict["restored_records"] = counter

            return (
                kvstore_collections_global_records_to_be_restored,
                kvstore_collections_global_records_restored,
                kvstore_collection_restore_summary_dict,
            )

        describe = False

        # init
        dry_run = True

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False

            if not describe:
                # Get the mandatory backup_archive name
                try:
                    backup_archive = resp_dict["backup_archive"]
                except Exception as e:
                    response = {
                        "response": "The backup_archive name is required in the body of the request"
                    }
                    logging.error(json.dumps(response, indent=2))
                    return {"payload": response, "status": 500}

                # Dry run mode, preview and verify the restore operation, but do not do anything
                # default to True
                try:
                    dry_run = resp_dict["dry_run"]
                    if dry_run in ("True", "true", "1", 1):
                        dry_run = True
                    elif dry_run in ("False", "false", "0", 0):
                        dry_run = False
                    else:
                        return {
                            "payload": {
                                "response": f'Invalid value for dry_run, received="{dry_run}", expecting a boolean value or true / false'
                            },
                            "status": 500,
                        }

                except Exception as e:
                    dry_run = True

                #
                # restore_knowledge_objects (False/True)
                #

                try:
                    restore_knowledge_objects = resp_dict["restore_knowledge_objects"]
                    if restore_knowledge_objects in ("True", "true", "1", 1):
                        restore_knowledge_objects = True
                    else:
                        restore_knowledge_objects = False
                except Exception as e:
                    restore_knowledge_objects = True

                #
                # knowledge_objects_tenants_scope (all / comma separated list of tenant_id)
                #

                try:
                    knowledge_objects_tenants_scope = resp_dict[
                        "knowledge_objects_tenants_scope"
                    ]
                    if knowledge_objects_tenants_scope == "all":
                        knowledge_objects_tenants_scope = "all"
                    else:
                        if not isinstance(knowledge_objects_tenants_scope, list):
                            knowledge_objects_tenants_scope = (
                                knowledge_objects_tenants_scope.split(",")
                            )
                        # for each entry in the list, strip and remove any empty char before or after the list entry
                        knowledge_objects_tenants_scope = [
                            x.strip() for x in knowledge_objects_tenants_scope
                        ]
                except Exception as e:
                    knowledge_objects_tenants_scope = "all"

                #
                # knowledge_objects_lists (all / comma separated list of objects to be restored)
                #

                try:
                    knowledge_objects_lists = resp_dict["knowledge_objects_lists"]
                    if knowledge_objects_lists == "all":
                        knowledge_objects_lists = "all"
                    else:
                        if not isinstance(knowledge_objects_lists, list):
                            knowledge_objects_lists = knowledge_objects_lists.split(",")
                        # for each entry in the list, strip and remove any empty char before or after the list entry
                        knowledge_objects_lists = [
                            x.strip() for x in knowledge_objects_lists
                        ]
                except Exception as e:
                    knowledge_objects_lists = "all"

                #
                # knowledge_objects_replace_existing (False/True) - If the object already exists, it will be replaced
                #

                try:
                    knowledge_objects_replace_existing = resp_dict[
                        "knowledge_objects_replace_existing"
                    ]
                    if knowledge_objects_replace_existing in ("True", "true", "1", 1):
                        knowledge_objects_replace_existing = True
                    else:
                        knowledge_objects_replace_existing = False
                except Exception as e:
                    knowledge_objects_replace_existing = True

                #
                # restore_kvstore_collections (False/True)
                #

                try:
                    restore_kvstore_collections = resp_dict[
                        "restore_kvstore_collections"
                    ]
                    if restore_kvstore_collections in ("True", "true", "1", 1):
                        restore_kvstore_collections = True
                    else:
                        restore_kvstore_collections = False
                except Exception as e:
                    restore_kvstore_collections = True

                #
                # kvstore_collections_scope (all / comma separated list of KVstore collections)
                #

                try:
                    kvstore_collections_scope = resp_dict["kvstore_collections_scope"]
                    if kvstore_collections_scope == "all":
                        kvstore_collections_scope = "all"
                    else:
                        if not isinstance(kvstore_collections_scope, list):
                            kvstore_collections_scope = kvstore_collections_scope.split(
                                ","
                            )
                        # for each entry in the list, strip and remove any empty char before or after the list entry
                        kvstore_collections_scope = [
                            x.strip() for x in kvstore_collections_scope
                        ]
                except Exception as e:
                    kvstore_collections_scope = "all"

                #
                # kvstore_collections_clean_empty (False/True) - If the collection was empty in the backup, restoring will empty any existing record in the collection
                #

                try:
                    kvstore_collections_clean_empty = resp_dict[
                        "kvstore_collections_clean_empty"
                    ]
                    if kvstore_collections_clean_empty in ("True", "true", "1", 1):
                        kvstore_collections_clean_empty = True
                    else:
                        kvstore_collections_clean_empty = False
                except Exception as e:
                    kvstore_collections_clean_empty = True

                #
                # kvstore_collections_restore_non_tenants_collections (False/True) - Restore non-tenants collections
                #

                try:
                    kvstore_collections_restore_non_tenants_collections = resp_dict[
                        "kvstore_collections_restore_non_tenants_collections"
                    ]
                    if kvstore_collections_restore_non_tenants_collections in (
                        "True",
                        "true",
                        "1",
                        1,
                    ):
                        kvstore_collections_restore_non_tenants_collections = True
                    else:
                        kvstore_collections_restore_non_tenants_collections = False
                except Exception as e:
                    kvstore_collections_restore_non_tenants_collections = True

                #
                # restore_virtual_tenant_accounts (False/True) - Restore virtual tenant accounts
                #

                try:
                    restore_virtual_tenant_accounts = resp_dict[
                        "restore_virtual_tenant_accounts"
                    ]
                    if restore_virtual_tenant_accounts in ("True", "true", "1", 1):
                        restore_virtual_tenant_accounts = True
                    else:
                        restore_virtual_tenant_accounts = False

                except Exception as e:
                    restore_virtual_tenant_accounts = True

                #
                # restore_virtual_tenant_main_kvrecord (False/True) - Restore virtual tenant main record
                #

                try:
                    restore_virtual_tenant_main_kvrecord = resp_dict[
                        "restore_virtual_tenant_main_kvrecord"
                    ]
                    if restore_virtual_tenant_main_kvrecord in ("True", "true", "1", 1):
                        restore_virtual_tenant_main_kvrecord = True
                    else:
                        restore_virtual_tenant_main_kvrecord = False

                except Exception as e:
                    restore_virtual_tenant_main_kvrecord = True

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint can be used to restore TrackMe Knowledge Objects and/or KVstore collections from a backup archive, the archive must be a tarball compressed made available in the backup directory of the application. ($SPLUNK_HOME/etc/apps/trackme/backup) - Restore operations for both knowledge objects and KVstore records are fully based on Splunk API usage, therefore compatible indifferently with Splunk Enterprise and Splunk Cloud, in a standalone or Search Head Cluster context. It requires a POST call with the following arguments:",
                "resource_desc": "Restore TrackMe Knowledge Objects and/or KVstore collections from a backup archive",
                "resource_spl_example": "| trackme url=\"/services/trackme/v2/backup_and_restore/restore\" mode=\"post\" body=\"{'backup_archive': 'trackme-backup-20210205-142635.tgz', 'dry_run': 'false', 'target': 'all'}\"",
                "options": [
                    {
                        "dry_run": "(true / false) OPTIONAL: if true, the endpoint will only verify that the archive can be found and successfully extracted, there will be no modifications at all. (default to true)",
                        "restore_virtual_tenant_accounts": "(true / false) OPTIONAL: check and restore the virtual tenant accounts from the submitted archive (default to true)",
                        "restore_virtual_tenant_main_kvrecord": "(true / false) OPTIONAL: check and restore the virtual tenant main record from the submitted archive (default to true)",
                        "restore_knowledge_objects": "(true / false) OPTIONAL: restore the knowledge objects from the submitted archive (default to true) - This requires a backup archive generated with TrackMe 2.1.5 or later",
                        "knowledge_objects_tenants_scope": "(all / comma separated list of tenant_id) OPTIONAL: restore the knowledge objects for all tenants or provide a comma separated list of tenant_id to be restored (defaults to all) - This requires a backup archive generated with TrackMe 2.1.5 or later",
                        "knowledge_objects_lists": "(all / comma separated list of objects to be restored) OPTIONAL: restore all knowledge objects that were backed up in the submitted archive, or provide a comma separated list of objects to be restored (defaults to all) - This requires a backup archive generated with TrackMe 2.1.5 or later",
                        "knowledge_objects_replace_existing": "(true / false) OPTIONAL: if the object already exists, it will be replaced (default to true) - This requires a backup archive generated with TrackMe 2.1.5 or later",
                        "restore_kvstore_collections": "(true / false) OPTIONAL: restore the KVstore collections from the submitted archive (default to true)",
                        "kvstore_collections_scope": "(all / comma separated list of KVstore collections) OPTIONAL: restore all collections that were backed up in the submitted archive, or provide a comma separated list of collections to be restored (defaults to all)",
                        "kvstore_collections_clean_empty": "(true / false) OPTIONAL: if the collection was empty in the backup, restoring will empty any existing record in the collection (default to true)",
                        "kvstore_collections_restore_non_tenants_collections": "(true / false) OPTIONAL: restore non-tenants collections (default to true), non tenants collections are KVstore collections which are tenant specific, such as user preferences, user settings, etc. If set to True, the content of the collection content will be restored from the backup.",
                        "backup_archive": "The archive file to be restored, the tarball compressed file must be located in the backup directory of the trackMe application.",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        else:
            # Get splunkd port
            splunkd_port = request_info.server_rest_port

            # Get service
            service = client.connect(
                owner="nobody",
                app="trackme",
                port=splunkd_port,
                token=request_info.system_authtoken,
                timeout=300,
            )

            # header
            header = {
                "Authorization": "Splunk %s" % request_info.session_key,
                "Content-Type": "application/json",
            }

            # set loglevel
            loglevel = trackme_getloglevel(
                request_info.system_authtoken, request_info.server_rest_port
            )
            log.setLevel(logging.getLevelName(loglevel))

            # Set backup root dir
            backuproot = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")

            # Set the submitted full path of the archive file
            backupfile = os.path.join(backuproot, backup_archive)

            # Set the full path of the directory for the extraction
            backupdir = os.path.splitext(backupfile)[0]

            # A dict for kvstore records summary
            kvstore_collections_restore_summary_dict = {}

            #################
            # Task: log start
            #################

            logging.info(
                f'TrackMe restore process is starting requested by user="{request_info.user}", request="{json.dumps(resp_dict, indent=3)}"'
            )

            # log the value for each argument
            logging.info(
                f"TrackMe restore process, listing arguments summary, "
                f'backup_archive="{backup_archive}", dry_run="{dry_run}", '
                f'restore_virtual_tenant_accounts="{restore_virtual_tenant_accounts}", '
                f'restore_virtual_tenant_main_kvrecord="{restore_virtual_tenant_main_kvrecord}", '
                f'restore_knowledge_objects="{restore_knowledge_objects}", '
                f'knowledge_objects_tenants_scope="{knowledge_objects_tenants_scope}", knowledge_objects_lists="{knowledge_objects_lists}", '
                f'knowledge_objects_replace_existing="{knowledge_objects_replace_existing}", restore_kvstore_collections="{restore_kvstore_collections}", '
                f'kvstore_collections_scope="{kvstore_collections_scope}", kvstore_collections_clean_empty="{kvstore_collections_clean_empty}", '
                f'kvstore_collections_restore_non_tenants_collections="{kvstore_collections_restore_non_tenants_collections}"'
            )

            ##########################
            # Task: Archive Extraction
            ##########################

            # First, check the backup archive existence
            if not os.path.isfile(backupfile):
                response = {
                    "response": f"The archive name {backupfile} could not be found on the file-system, restore cannot be processed",
                }
                logging.error(json.dumps(response, indent=2))
                return {"payload": response, "status": 500}

            # Attempt extraction
            try:
                tf = tarfile.open(backupfile)
                tf.extractall(backupdir)
            except Exception as e:
                response = {
                    "response": f"The archive name {backupfile} could not be extracted, restore cannot be processed, exception='{str(e)}'",
                }
                logging.error(json.dumps(response, indent=2))
                return {"payload": response, "status": 500}

            ######################
            # Task: Check metadata
            ######################

            # default to 2.0.0 (metadata compatibility, generated with TrackMe >= 2.1.5)
            archive_schema_version = "2.0.0"

            # check metadata files
            full_metadata_file = f"{backupfile}.full.meta"
            light_metadata_file = f"{backupfile}.light.meta"

            # From TrackMe 2.1.5, we generate a metadata file named as <tarfile>.full.meta and <tarfile>.light.meta
            # Check if the metadata file exists
            backup_file_has_metadata = False

            # check for both metadata files

            if os.path.isfile(full_metadata_file) and os.path.isfile(
                light_metadata_file
            ):

                backup_file_has_metadata = True

                # full metadata
                try:
                    with open(f"{backupfile}.full.meta", "r") as read_content:
                        full_metadata = json.load(read_content)
                except Exception as e:
                    full_metadata = None
                    backup_file_has_metadata = False
                    logging.error(
                        f'failed to load the full metadata file="{backupfile}.full.meta" with exception="{str(e)}"'
                    )

                # light metadata
                try:
                    with open(f"{backupfile}.light.meta", "r") as read_content:
                        light_metadata = json.load(read_content)
                except Exception as e:
                    light_metadata = None
                    backup_file_has_metadata = False
                    logging.error(
                        f'failed to load the light metadata file="{backupfile}.light.meta" with exception="{str(e)}"'
                    )

                # if we do not have full metadata or light metadata, the archive schema version is 1.0.0 (Only KVstore collections can be restored)
                if not full_metadata or not light_metadata:
                    archive_schema_version = "1.0.0"

            ############################################
            # Task: Parse restorable KVstore collections
            ############################################

            # store the list of KVstore collection json files in a list
            collections_json_files = [
                f
                for f in listdir(backupdir)
                if isfile(join(backupdir, f)) and f.startswith("kv_")
            ]

            # store the list of available collections in the archive
            collections_available = []

            # create a dictionnary
            collections_restore_dict = {}
            for json_file in collections_json_files:
                # strip the extension
                collection_name = os.path.splitext(json_file)[0]

                # append to the list of available collections for restore
                collections_available.append(collection_name)

                # get the file size
                json_file_size = os.path.getsize(os.path.join(backupdir, json_file))

                # get the file mtime
                json_file_mtime = round(
                    os.path.getmtime(os.path.join(backupdir, json_file))
                )

                # try getting the number of records
                try:
                    with open(os.path.join(backupdir, json_file), "r") as read_content:
                        json_file_records = len(json.load(read_content))
                except Exception as e:
                    json_file_records = None

                # add to the dict
                is_empty = False
                if json_file_records == 0:
                    is_empty = True

                collections_restore_dict[collection_name] = {
                    "file": json_file,
                    "size": json_file_size,
                    "mtime": json_file_mtime,
                    "records": json_file_records,
                    "is_empty": is_empty,
                }

            ##############
            # dry_run mode
            ##############

            # if dry run
            if dry_run:

                # init dry_run response
                dry_run_response = {}

                # add response
                response_message = f"Success, the archive {backupfile} could be successfully extracted, consult metadata information added to this response for more insights. You can run this command with dry_run=false to perform the restoration."
                dry_run_response["response"] = response_message

                # if metadata files are available, add the light_metadata as metadata
                if backup_file_has_metadata:
                    dry_run_response["knowledge_objects_summary"] = light_metadata.get(
                        "knowledge_objects_summary"
                    )
                    dry_run_response["metadata"] = light_metadata

                # add collections
                dry_run_response["kvstore_collections_json_files"] = (
                    collections_json_files
                )
                dry_run_response["kvstore_collections_details"] = (
                    collections_restore_dict
                )

                # iterate through knowledge objects files and add to the response_message, as well as the virtual tenant account
                knowledge_objects_json_files = [
                    f
                    for f in listdir(backupdir)
                    if isfile(join(backupdir, f))
                    and f.startswith("tenant_")
                    and f.endswith("_knowledge_objects.json")
                ]

                for knowledge_objects_json_file in knowledge_objects_json_files:

                    # extract the tenant_id
                    tenant_id = knowledge_objects_json_file.split("_")[1]

                    if (
                        knowledge_objects_tenants_scope == "all"
                        or tenant_id in knowledge_objects_tenants_scope
                    ):

                        # open and load the json file
                        try:
                            with open(
                                os.path.join(backupdir, knowledge_objects_json_file),
                                "r",
                            ) as read_content:
                                knowledge_objects_json = json.load(read_content)
                        except Exception as e:
                            clean_backup_dir(backupdir)
                            knowledge_objects_json = None
                            logging.error(
                                f'failed to load the knowledge_objects_json file="{knowledge_objects_json_file}" with exception="{str(e)}"'
                            )

                        # iterate through the knowledge objects lists, add each category of objects to the response
                        # categories: savedsearches, alerts, macros, lookup_definitions, kvstore_collections
                        savedsearches_list = []
                        alerts_list = []
                        macros_list = []
                        lookup_definitions_list = []
                        kvstore_collections_list = []

                        # Iterate through the values of the dictionary
                        for ko in knowledge_objects_json.values():
                            type_ko = ko.get("type")
                            if type_ko == "savedsearches":
                                savedsearches_list.append(ko)
                            elif type_ko == "alerts":
                                alerts_list.append(ko)
                            elif type_ko == "macros":
                                macros_list.append(ko)
                            elif type_ko == "lookup_definitions":
                                lookup_definitions_list.append(ko)
                            elif type_ko == "kvstore_collections":
                                kvstore_collections_list.append(ko)

                        # Add to the response
                        dry_run_response[f"tenant_{tenant_id}_knowledge_objects"] = {
                            "savedsearches": savedsearches_list,
                            "alerts": alerts_list,
                            "macros": macros_list,
                            "lookup_definitions": lookup_definitions_list,
                            "kvstore_collections": kvstore_collections_list,
                        }

                        # check that the virtual account backup file exists, if not log and skip
                        tenant_json_file = f"tenant_{tenant_id}_vtenant_account.json"

                        if not os.path.isfile(
                            os.path.join(backupdir, tenant_json_file)
                        ):
                            clean_backup_dir(backupdir)
                            error_message = f'failed to find the backup file="{tenant_json_file}" for tenant_id="{tenant_id}", the tenant cannot be restored'
                            logging.error(error_message)
                            return {
                                "payload": {
                                    "response": error_message,
                                },
                                "status": 500,
                            }

                        # load the tenant json file
                        try:
                            f = open(os.path.join(backupdir, tenant_json_file), "r")
                            vtenant_account_data = json.loads(f.read())
                            info_message = f'loaded the tenant backup file="{tenant_json_file}" for tenant_id="{tenant_id}"'
                            logging.info(info_message)

                        except Exception as e:
                            clean_backup_dir(backupdir)
                            error_message = f'failed to open json file="{os.path.join(backupdir, tenant_json_file)}" for reading with exception="{str(e)}", cannot restore tenant_id="{tenant_id}"'
                            logging.error(error_message)
                            return {
                                "payload": {
                                    "response": error_message,
                                },
                                "status": 500,
                            }

                        # add to the response
                        dry_run_response[f"tenant_{tenant_id}_vtenant_account"] = (
                            vtenant_account_data
                        )

                # log and render

                # remove backup dir
                clean_backup_dir(backupdir)
                logging.info(response_message)
                return {"payload": dry_run_response, "status": 200}

            else:

                ###################
                # Live restore mode
                ###################

                # create a dict to store the restore results
                restore_results_dict = {}

                # create global counters
                kvstore_collections_global_records_to_be_restored = 0
                kvstore_collections_global_records_restored = 0

                # store collections with failures in a specific list
                kvstore_collections_restored_warning = []

                # Things are serious now, let's restore collection per collection

                # set the list of collections to be restored
                kvstore_collections_to_be_restored = []

                # set the list of collections that were restored
                kvstore_collections_restored = []

                # if kvstore_collections_scope is all, add every collection that can be restored
                if kvstore_collections_scope == "all":
                    # Loop through the collections
                    for collection_name in collections_restore_dict:
                        kvstore_collections_to_be_restored.append(collection_name)

                else:
                    # Loop and check if the requested collection is available for restore
                    for collection_name in kvstore_collections_scope:
                        if not collection_name in collections_available:
                            # an impossible operaton is requested, we cannot proceed
                            response = {
                                "action": "failure",
                                "response": f'the collection="{collection_name}" requested for restoration is not available in the backup archive file, restore cannot be procceded',
                                "collections_available": collections_available,
                            }

                            return {"payload": response, "status": 500}

                        else:
                            kvstore_collections_to_be_restored.append(collection_name)

                #########
                # Proceed
                #########

                # init final response_dict
                response_dict = {}

                # init tasks list
                tasks_list = []

                # init errors list
                errors_list = []

                #####################################
                # Task: Restore the Knowledge Objects
                #####################################

                if archive_schema_version == "2.0.0" and restore_knowledge_objects:
                    logging.info(
                        f'Detected archive_schema_version="{archive_schema_version}", Processing restore operations of Knowledge Objects'
                    )

                    # knowledge_objects_json_files: list of json files in the backup archive
                    # each file is named as:
                    # tenant_<tenant_id>_knowledge_objects.json

                    tenants_id_restorable = []
                    tenants_id_available = []

                    knowledge_objects_json_files = [
                        f
                        for f in listdir(backupdir)
                        if isfile(join(backupdir, f))
                        and f.startswith("tenant_")
                        and f.endswith("_knowledge_objects.json")
                    ]

                    for knowledge_objects_json_file in knowledge_objects_json_files:
                        # extract the tenant_id
                        tenant_id = knowledge_objects_json_file.split("_")[1]

                        # add to available
                        tenants_id_available.append(tenant_id)

                        if knowledge_objects_tenants_scope == "all":
                            tenants_id_restorable.append(tenant_id)
                        else:
                            if tenant_id in knowledge_objects_tenants_scope:
                                tenants_id_restorable.append(tenant_id)

                    # if a specific tenant_id was requested, but not found in the backup archive, stop and return an error
                    if knowledge_objects_tenants_scope != "all":
                        for tenant_id in knowledge_objects_tenants_scope:
                            if tenant_id not in tenants_id_restorable:
                                error_message = f'failed to find the backup file for tenant_id="{tenant_id}", the tenant cannot be restored, verify that your input is valid.'
                                logging.error(error_message)
                                return {
                                    "payload": {
                                        "response": error_message,
                                        "tenants_id_available": tenants_id_available,
                                    },
                                    "status": 500,
                                }

                    ###############################################
                    # subtask: load the central KVstore from backup
                    ###############################################

                    collection_name_main = "kv_trackme_virtual_tenants"

                    # connect to live KVstore collection
                    try:
                        collection = service.kvstore[collection_name_main]
                    except Exception as e:
                        error_message = f'failed to connect to KVstore collection="{collection_name_main}" with exception="{str(e)}"'
                        logging.error(error_message)
                        return {
                            "payload": {
                                "response": error_message,
                            },
                            "status": 500,
                        }

                    # backup file name
                    source_json_file = collections_restore_dict[collection_name_main][
                        "file"
                    ]

                    # try loading the json data
                    try:
                        f = open(os.path.join(backupdir, source_json_file), "r")
                        vtenants_main_records = json.loads(f.read())

                    except Exception as e:
                        error_message = f'failed to open json file="{os.path.join(backupdir, source_json_file)}" for reading with exception="{str(e)}"'
                        logging.error(error_message)
                        return {
                            "payload": {
                                "response": error_message,
                            },
                            "status": 500,
                        }

                    ###################################################
                    # subtask: restore per tenant the knowledge objects
                    ###################################################

                    # Loop through the restorable tenants, and proceed
                    for tenant_id in tenants_id_restorable:
                        logging.info(
                            f'Processing with restore operations for tenant_id="{tenant_id}"'
                        )

                        #############################################
                        # subtask: restore the Virtual Tenant account
                        #############################################

                        # check that the virtual account backup file exists, if not log and skip
                        tenant_json_file = f"tenant_{tenant_id}_vtenant_account.json"

                        if not os.path.isfile(
                            os.path.join(backupdir, tenant_json_file)
                        ):
                            error_message = f'failed to find the backup file="{tenant_json_file}" for tenant_id="{tenant_id}", the tenant cannot be restored'
                            logging.error(error_message)
                            errors_list.append(error_message)
                            break

                        # load the tenant json file
                        try:
                            f = open(os.path.join(backupdir, tenant_json_file), "r")
                            data = json.loads(f.read())
                            info_message = f'loaded the tenant backup file="{tenant_json_file}" for tenant_id="{tenant_id}"'
                            logging.info(info_message)
                            tasks_list.append(info_message)

                        except Exception as e:
                            error_message = f'failed to open json file="{os.path.join(backupdir, tenant_json_file)}" for reading with exception="{str(e)}", cannot restore tenant_id="{tenant_id}"'
                            logging.error(error_message)
                            errors_list.append(error_message)
                            break

                        #
                        # Proceed
                        #

                        if not restore_virtual_tenant_accounts:
                            info_message = f'restore_virtual_tenant_accounts is set to False, the Virtual Tenant account for tenant_id="{tenant_id}" will not be restored'
                            logging.info(info_message)
                            tasks_list.append(info_message)

                        else:

                            # If the Virtual Account exists, it must be deleted to be restored

                            # Del the vtenant account
                            url = "%s/servicesNS/nobody/trackme/trackme_vtenants/%s" % (
                                request_info.server_rest_uri,
                                tenant_id,
                            )

                            # Retrieve and set the tenant idx, if any failure, logs and use the global index
                            try:
                                response = requests.delete(
                                    url, headers=header, verify=False, timeout=300
                                )
                                if response.status_code not in (200, 201, 204):
                                    # this can be expected if the tenant does not exist, do not consider this as an error
                                    info_message = f'tenant_id="{tenant_id}", delete vtenant account was not required, response.status_code="{response.status_code}"'
                                    logging.info(info_message)
                                    tasks_list.append(info_message)
                                else:
                                    info_message = f'deleted the Virtual Tenant account for tenant_id="{tenant_id}"'
                                    logging.info(info_message)
                                    tasks_list.append(info_message)
                            except Exception as e:
                                # this is an error
                                error_message = f'failed to delete the Virtual Tenant account for tenant_id="{tenant_id}" with exception="{str(e)}"'
                                logging.error(error_message)
                                errors_list.append(error_message)

                            # Remove keys that start with "eai:"
                            keys_to_remove = [
                                k for k in data.keys() if k.startswith("eai:")
                            ]
                            for k in keys_to_remove:
                                del data[k]

                            # also remove disabled
                            if "disabled" in data:
                                del data["disabled"]

                            # in data, add the key name equal to the tenant_id
                            data["name"] = tenant_id

                            # Add the vtenant account
                            url = "%s/servicesNS/nobody/trackme/trackme_vtenants" % (
                                request_info.server_rest_uri
                            )

                            # Run the POST call
                            try:

                                response = requests.post(
                                    url,
                                    headers=header,
                                    data=data,
                                    verify=False,
                                    timeout=300,
                                )
                                response.raise_for_status()
                                info_message = f'added the Virtual Tenant account for tenant_id="{tenant_id}"'
                                logging.info(info_message)
                                tasks_list.append(info_message)

                            except Exception as e:
                                error_message = f'failed to add the Virtual Tenant account for tenant_id="{tenant_id}" with exception="{str(e)}"'
                                logging.error(error_message)
                                errors_list.append(error_message)
                                break

                        #####################################################
                        # subtask: restore the central record for this tenant
                        #####################################################

                        #
                        # Proceed
                        #

                        if not restore_virtual_tenant_main_kvrecord:

                            info_message = f'restore_virtual_tenant_main_kvrecord is set to False, the main KVstore record for tenant_id="{tenant_id}" will not be restored'
                            logging.info(info_message)
                            tasks_list.append(info_message)

                        else:

                            tenant_id_main_record = None
                            for record in vtenants_main_records:
                                if record["tenant_id"] == tenant_id:
                                    tenant_id_main_record = record
                                    break

                            if not tenant_id_main_record:
                                error_message = f'failed to find the tenant_id="{tenant_id}" in the backup file="{source_json_file}", the tenant cannot be restored'
                                logging.error(error_message)
                                errors_list.append(error_message)
                                break
                            else:
                                logging.info(
                                    f'Found tenant_id="{tenant_id}" in the backup file="{source_json_file}", the tenant can be restored'
                                )

                            # in live KVstore, check if the tenant_id already exists, it it does, replace the record, otherwise create it
                            try:
                                existing_record = collection.data.query(
                                    query=json.dumps({"tenant_id": tenant_id})
                                )[0]
                            except Exception as e:
                                existing_record = None

                            # if the tenant_id already exists, replace the record
                            if existing_record:
                                try:
                                    collection.data.update(
                                        existing_record["_key"],
                                        json.dumps(tenant_id_main_record),
                                    )
                                    tasks_list.append(
                                        f'updated the tenant_id="{tenant_id}" in the KVstore collection="{collection_name_main}"'
                                    )
                                except Exception as e:
                                    error_message = f'failed to update the tenant_id="{tenant_id}" in the KVstore collection="{collection_name_main}" with exception="{str(e)}"'
                                    logging.error(error_message)
                                    return {
                                        "payload": {
                                            "response": error_message,
                                        },
                                        "status": 500,
                                    }
                            else:
                                try:
                                    collection.data.insert(
                                        json.dumps(tenant_id_main_record)
                                    )
                                    tasks_list.append(
                                        f'inserted the tenant_id="{tenant_id}" in the KVstore collection="{collection_name_main}"'
                                    )
                                except Exception as e:
                                    error_message = f'failed to insert the tenant_id="{tenant_id}" in the KVstore collection="{collection_name_main}" with exception="{str(e)}"'
                                    logging.error(error_message)
                                    return {
                                        "payload": {
                                            "response": error_message,
                                        },
                                        "status": 500,
                                    }

                        ##########################################################
                        # task: load knowledge objects backup file for this tenant
                        ##########################################################

                        knowledge_objects_json_file = (
                            f"tenant_{tenant_id}_knowledge_objects.json"
                        )
                        try:
                            f = open(
                                os.path.join(backupdir, knowledge_objects_json_file),
                                "r",
                            )
                            data = json.loads(f.read())
                            info_message = f'loaded the knowledge objects backup file="{knowledge_objects_json_file}" for tenant_id="{tenant_id}"'
                            logging.info(info_message)
                            tasks_list.append(info_message)
                        except Exception as e:
                            error_message = f'failed to open json file="{os.path.join(backupdir, knowledge_objects_json_file)}" for reading with exception="{str(e)}", cannot restore knowledge objects for tenant_id="{tenant_id}"'
                            logging.error(error_message)
                            errors_list.append(error_message)
                            break

                        # init lists and dicts of knowledge objects for this tenant in the backup data
                        kvstore_collections_restorable_list = []
                        kvstore_collections_restorable_dict = {}
                        kvstore_transforms_restorable_list = []
                        kvstore_transforms_restorable_dict = {}
                        macros_restorable_list = []
                        macros_restorable_dict = {}
                        savedsearches_restorable_list = []
                        savedsearches_restorable_dict = {}
                        alerts_restorable_list = []
                        alerts_restorable_dict = {}

                        # first parse the data, get the type of object and the name, add to the appropriate list, log and add to the tasks list
                        for record in data.values():

                            if record["type"] == "kvstore_collections":
                                kvstore_collections_restorable_list.append(
                                    record["title"]
                                )
                                kvstore_collections_restorable_dict[record["title"]] = (
                                    record
                                )
                            if record["type"] == "lookup_definitions":
                                kvstore_transforms_restorable_list.append(
                                    record["title"]
                                )
                                kvstore_transforms_restorable_dict[record["title"]] = (
                                    record
                                )
                            if record["type"] == "macros":
                                macros_restorable_list.append(record["title"])
                                macros_restorable_dict[record["title"]] = record
                            if record["type"] == "savedsearches":
                                savedsearches_restorable_list.append(record["title"])
                                savedsearches_restorable_dict[record["title"]] = record
                            if record["type"] == "alerts":
                                alerts_restorable_list.append(record["title"])
                                alerts_restorable_dict[record["title"]] = record

                        info_message = f'knowledge objects backup file="{knowledge_objects_json_file}" for tenant_id="{tenant_id}" parsed successfully, number of restorable kvstore collections="{len(kvstore_collections_restorable_list)}", number of restorable transforms definitions="{len(kvstore_transforms_restorable_list)}", number of restorable macros="{len(macros_restorable_list)}", number of restorable saved searches="{len(savedsearches_restorable_list)}"'
                        logging.info(info_message)
                        tasks_list.append(info_message)

                        # add to response_dict
                        response_dict[f"knowledge_objects_tenant_id_{tenant_id}"] = {
                            "kvstore_collections": kvstore_collections_restorable_list,
                            "transforms_definitions": kvstore_transforms_restorable_list,
                            "macros": macros_restorable_list,
                            "savedsearches": savedsearches_restorable_list,
                            "alerts": alerts_restorable_list,
                        }

                        ##########################################################
                        # task: restore kvstore collections definitions for tenant
                        ##########################################################

                        # Loop through the restorable collections, and proceed
                        if restore_kvstore_collections:
                            for collection_name in kvstore_collections_restorable_list:
                                logging.info(
                                    f'TrackMe restore process starting, knowledge objects, processing restore of kvstore definition for collection_name="{collection_name}" for tenant_id="{tenant_id}"'
                                )

                                # check if the collection exists already
                                collection_exists = False

                                try:
                                    collection = service.kvstore[collection_name]
                                    collection_exists = True

                                except Exception as e:
                                    pass

                                # if knowledge_objects_replace_existing, systematically attempt to delete the collection
                                if (
                                    collection_exists
                                    and not knowledge_objects_replace_existing
                                ):

                                    logging.info(
                                        f'collection="{collection_name}" already exists and the replace_existing option is set to False, skipping the restore operation'
                                    )
                                    result = {
                                        "object": collection_name,
                                        "object_type": "kvstore collection",
                                        "action": "restore",
                                        "result": "skipped",
                                        "reason": "already exists and replace_existing is set to False",
                                    }
                                    logging.info(json.dumps(result, indent=4))
                                    tasks_list.append(result)
                                    continue

                                else:

                                    if collection_exists:
                                        try:
                                            action = trackme_delete_kvcollection(
                                                request_info.system_authtoken,
                                                request_info.server_rest_uri,
                                                tenant_id,
                                                collection_name,
                                            )
                                            logging.info(
                                                f'collection="{collection_name}" was deleted successfully, response="{action}"'
                                            )
                                            result = {
                                                "object": collection_name,
                                                "object_type": "kvstore collection",
                                                "action": "delete",
                                                "result": "success",
                                            }
                                            tasks_list.append(result)
                                        except Exception as e:
                                            pass  # no need to log or report, the collection might not exist yet

                                try:

                                    ko_acl = {
                                        "owner": kvstore_collections_restorable_dict[
                                            collection_name
                                        ]["properties"]["eai:acl.owner"],
                                        "sharing": kvstore_collections_restorable_dict[
                                            collection_name
                                        ]["properties"]["eai:acl.sharing"],
                                        "perms.write": kvstore_collections_restorable_dict[
                                            collection_name
                                        ][
                                            "properties"
                                        ][
                                            "eai:acl.perms.write"
                                        ],
                                        "perms.read": kvstore_collections_restorable_dict[
                                            collection_name
                                        ][
                                            "properties"
                                        ][
                                            "eai:acl.perms.read"
                                        ],
                                    }

                                    trackme_create_kvcollection(
                                        request_info.system_authtoken,
                                        request_info.server_rest_uri,
                                        tenant_id,
                                        collection_name,
                                        ko_acl,
                                    )
                                    result = {
                                        "object": collection_name,
                                        "object_type": "kvstore collection",
                                        "action": "restore",
                                        "result": "success",
                                    }
                                    tasks_list.append(result)
                                except Exception as e:
                                    result = {
                                        "object": collection_name,
                                        "object_type": "kvstore collection",
                                        "action": "restore",
                                        "result": "failure",
                                        "exception": str(e),
                                    }
                                    errors_list.append(result)

                                # restore the collection records for this collection and add to the list of restored collections
                                (
                                    kvstore_collections_global_records_to_be_restored,
                                    kvstore_collections_global_records_restored,
                                    kvstore_collection_restore_summary_dict,
                                ) = restore_kvstore_records(
                                    service,
                                    collection_name,
                                    collections_restore_dict,
                                    backupdir,
                                    kvstore_collections_global_records_to_be_restored,
                                    kvstore_collections_global_records_restored,
                                    kvstore_collections_restored_warning,
                                    restore_results_dict,
                                    kvstore_collections_clean_empty,
                                )

                                kvstore_collections_restored.append(collection_name)

                                # add to dict
                                kvstore_collections_restore_summary_dict[
                                    collection_name
                                ] = kvstore_collection_restore_summary_dict

                        #########################################################
                        # task: restore kvstore transforms definitions for tenant
                        #########################################################

                        # Loop through the restorable collections transforms, and proceed

                        # if knowledge_objects_lists is not equal to all, filter the list to its value
                        if knowledge_objects_lists != "all":
                            kvstore_transforms_restorable_list = [
                                transform
                                for transform in kvstore_transforms_restorable_list
                                if transform in knowledge_objects_lists
                            ]

                        for transforms_name in kvstore_transforms_restorable_list:

                            logging.info(
                                f'TrackMe restore process starting, knowledge objects, processing restore of kvstore transforms definition for transforms_name="{transforms_name}" for tenant_id="{tenant_id}"'
                            )

                            # check if the transforms exists already
                            transforms_exists = False

                            try:
                                transforms_object = service.confs["transforms"][
                                    transforms_name
                                ]
                                transforms_exists = True

                            except Exception as e:
                                pass

                            if (
                                transforms_exists
                                and not knowledge_objects_replace_existing
                            ):
                                logging.info(
                                    f'transforms="{transforms_name}" already exists and the replace_existing option is set to False, skipping the restore operation'
                                )
                                result = {
                                    "object": transforms_name,
                                    "object_type": "kvstore transforms",
                                    "action": "restore",
                                    "result": "skipped",
                                    "reason": "already exists and replace_existing is set to False",
                                }
                                logging.info(json.dumps(result, indent=4))
                                tasks_list.append(result)
                                continue

                            else:

                                if transforms_exists:

                                    try:
                                        action = trackme_delete_kvtransform(
                                            request_info.system_authtoken,
                                            request_info.server_rest_uri,
                                            tenant_id,
                                            transforms_name,
                                        )
                                        result = {
                                            "object": transforms_name,
                                            "object_type": "transform",
                                            "action": "delete",
                                            "result": "success",
                                        }
                                        logging.info(json.dumps(result, indent=4))
                                        tasks_list.append(result)

                                    except Exception as e:
                                        result = {
                                            "object": transforms_name,
                                            "object_type": "transform",
                                            "action": "delete",
                                            "result": "failure",
                                            "exception": str(e),
                                        }
                                        logging.error(json.dumps(result, indent=4))
                                        errors_list.append(result)

                                try:

                                    ko_acl = {
                                        "owner": kvstore_transforms_restorable_dict[
                                            transforms_name
                                        ]["properties"]["eai:acl.owner"],
                                        "sharing": kvstore_transforms_restorable_dict[
                                            transforms_name
                                        ]["properties"]["eai:acl.sharing"],
                                        "perms.write": kvstore_transforms_restorable_dict[
                                            transforms_name
                                        ][
                                            "properties"
                                        ][
                                            "eai:acl.perms.write"
                                        ],
                                        "perms.read": kvstore_transforms_restorable_dict[
                                            transforms_name
                                        ][
                                            "properties"
                                        ][
                                            "eai:acl.perms.read"
                                        ],
                                    }

                                    trackme_create_kvtransform(
                                        request_info.system_authtoken,
                                        request_info.server_rest_uri,
                                        tenant_id,
                                        transforms_name,
                                        kvstore_transforms_restorable_dict[
                                            transforms_name
                                        ]["fields_list"],
                                        kvstore_transforms_restorable_dict[
                                            transforms_name
                                        ]["collection"],
                                        ko_acl.get("owner"),
                                        ko_acl,
                                    )
                                    result = {
                                        "object": transforms_name,
                                        "object_type": "kvstore transforms",
                                        "action": "restore",
                                        "result": "success",
                                    }
                                    logging.info(json.dumps(result, indent=4))
                                    tasks_list.append(result)

                                except Exception as e:
                                    result = {
                                        "object": transforms_name,
                                        "object_type": "kvstore transforms",
                                        "action": "restore",
                                        "result": "failure",
                                        "exception": str(e),
                                    }
                                    logging.error(json.dumps(result, indent=4))
                                    errors_list.append(result)

                        #############################################
                        # task: restore macros definitions for tenant
                        #############################################

                        # Loop through the restorable macros, and proceed

                        # if knowledge_objects_lists is not equal to all, filter the list to its value
                        if knowledge_objects_lists != "all":
                            macros_restorable_list = [
                                macro
                                for macro in macros_restorable_list
                                if macro in knowledge_objects_lists
                            ]

                        for macro_name in macros_restorable_list:

                            logging.info(
                                f'TrackMe restore process starting, knowledge objects, processing restore of macro definition for macro_name="{macro_name}" for tenant_id="{tenant_id}"'
                            )

                            # check if the macro exists already
                            macro_exists = False

                            try:
                                macro_object = service.confs["macros"][macro_name]
                                macro_exists = True
                            except Exception as e:
                                pass

                            if macro_exists and not knowledge_objects_replace_existing:
                                logging.info(
                                    f'macro="{macro_name}" already exists and the replace_existing option is set to False, skipping the restore operation'
                                )
                                result = {
                                    "object": macro_name,
                                    "object_type": "macro",
                                    "action": "restore",
                                    "result": "skipped",
                                    "reason": "already exists and replace_existing is set to False",
                                }
                                logging.info(json.dumps(result, indent=4))
                                tasks_list.append(result)
                                continue

                            else:

                                if macro_exists:

                                    try:
                                        action = trackme_delete_macro(
                                            request_info.system_authtoken,
                                            request_info.server_rest_uri,
                                            tenant_id,
                                            macro_name,
                                        )
                                        logging.info(
                                            f'macro="{macro_name}" was deleted successfully, response="{action}"'
                                        )
                                        result = {
                                            "object": macro_name,
                                            "object_type": "macro",
                                            "action": "delete",
                                            "result": "success",
                                        }
                                        logging.info(json.dumps(result, indent=4))
                                        tasks_list.append(result)

                                    except Exception as e:
                                        result = {
                                            "object": macro_name,
                                            "object_type": "macro",
                                            "action": "delete",
                                            "result": "failure",
                                            "exception": str(e),
                                        }
                                        logging.error(json.dumps(result, indent=4))
                                        errors_list.append(result)

                                try:

                                    ko_acl = {
                                        "owner": macros_restorable_dict[macro_name][
                                            "properties"
                                        ]["eai:acl.owner"],
                                        "sharing": macros_restorable_dict[macro_name][
                                            "properties"
                                        ]["eai:acl.sharing"],
                                        "perms.write": macros_restorable_dict[
                                            macro_name
                                        ]["properties"]["eai:acl.perms.write"],
                                        "perms.read": macros_restorable_dict[
                                            macro_name
                                        ]["properties"]["eai:acl.perms.read"],
                                    }

                                    trackme_create_macro(
                                        request_info.system_authtoken,
                                        request_info.server_rest_uri,
                                        tenant_id,
                                        macro_name,
                                        macros_restorable_dict[macro_name][
                                            "definition"
                                        ],
                                        ko_acl.get("owner"),
                                        ko_acl,
                                    )

                                    result = {
                                        "object": macro_name,
                                        "object_type": "macro",
                                        "action": "restore",
                                        "result": "success",
                                    }
                                    logging.info(json.dumps(result, indent=4))
                                    tasks_list.append(result)

                                except Exception as e:
                                    result = {
                                        "object": macro_name,
                                        "object_type": "macro",
                                        "action": "restore",
                                        "result": "failure",
                                        "exception": str(e),
                                    }
                                    logging.error(json.dumps(result, indent=4))
                                    errors_list.append(result)

                        ####################################################
                        # task: restore savedsearches definitions for tenant
                        ####################################################

                        # Save available savedsearches in a main list
                        savedsearches_restorable_list_origin = []
                        for savedsearch in savedsearches_restorable_dict:
                            savedsearches_restorable_list_origin.append(savedsearch)

                        # If knowledge_objects_lists is not equal to "all", filter the list to its value
                        if knowledge_objects_lists != "all":
                            savedsearches_restorable_list = [
                                savedsearch
                                for savedsearch in savedsearches_restorable_list_origin
                                if savedsearch in knowledge_objects_lists
                            ]
                        else:
                            savedsearches_restorable_list = (
                                savedsearches_restorable_list_origin.copy()
                            )

                        # Restoring saved searches has some challenges, if the search itself refers to another saved search, we need to
                        # restore this saved search first, and then restore the search that refers to it.

                        # Init a list of restored saved searches
                        savedsearches_restored = []

                        for savedsearch_name in savedsearches_restorable_list:

                            logging.info(
                                f'TrackMe restore process starting, knowledge objects, processing restore of savedsearch definition for savedsearch_name="{savedsearch_name}" for tenant_id="{tenant_id}"'
                            )

                            # First, get the search definition
                            savedsearch_definition = savedsearches_restorable_dict[
                                savedsearch_name
                            ]["definition"]

                            # Use regex to find the parent search
                            parent_search_match = re.search(
                                r"\| savedsearch\s+\"{0,1}([^\"]+)\"{0,1}",
                                savedsearch_definition,
                            )

                            # Extract parent search name if found
                            parent_search_name = (
                                parent_search_match.group(1)
                                if parent_search_match
                                else None
                            )

                            if parent_search_name:
                                # Check if the parent search exists in the original restorable list
                                if (
                                    parent_search_name
                                    in savedsearches_restorable_list_origin
                                ):
                                    # Ensure the parent search is restored first
                                    if parent_search_name not in savedsearches_restored:
                                        logging.info(
                                            f'TrackMe restore process: parent search "{parent_search_name}" detected for savedsearch "{savedsearch_name}". Restoring parent first.'
                                        )
                                        # Recursive restore call for the parent search
                                        if (
                                            parent_search_name
                                            in savedsearches_restorable_list
                                        ):
                                            savedsearches_restorable_list.remove(
                                                parent_search_name
                                            )  # Avoid duplicate processing
                                        savedsearches_restorable_list.insert(
                                            0, parent_search_name
                                        )  # Prioritize restoring the parent search
                                else:
                                    # Log an error and skip the child search if the parent isn't available
                                    error_message = {
                                        "object": savedsearch_name,
                                        "object_type": "savedsearch",
                                        "action": "restore",
                                        "result": "failure",
                                        "reason": f"Parent search '{parent_search_name}' not found in restorable list.",
                                    }
                                    logging.error(json.dumps(error_message, indent=4))
                                    errors_list.append(error_message)
                                    continue  # Skip restoring the current search

                            # Proceed with the restoration logic for the current saved search
                            savedsearch_exists = False
                            try:
                                savedsearch_object = service.saved_searches[
                                    savedsearch_name
                                ]
                                savedsearch_search = savedsearch_object.content[
                                    "search"
                                ]
                                savedsearch_exists = True

                            except Exception:
                                pass

                            if (
                                savedsearch_exists
                                and not knowledge_objects_replace_existing
                            ):
                                logging.info(
                                    f'savedsearch="{savedsearch_name}" already exists and the replace_existing option is set to False, skipping the restore operation'
                                )
                                result = {
                                    "object": savedsearch_name,
                                    "object_type": "savedsearch",
                                    "action": "restore",
                                    "result": "skipped",
                                    "reason": "already exists and replace_existing is set to False",
                                }
                                logging.info(json.dumps(result, indent=4))
                                tasks_list.append(result)
                            else:
                                # Restore logic
                                try:
                                    if savedsearch_exists:
                                        # Delete existing savedsearch if required
                                        action = trackme_delete_report(
                                            request_info.system_authtoken,
                                            request_info.server_rest_uri,
                                            tenant_id,
                                            savedsearch_name,
                                        )
                                        logging.info(
                                            f'savedsearch="{savedsearch_name}" was deleted successfully, response="{action}"'
                                        )

                                    # Prepare properties for restoration
                                    savedsearch_properties = {
                                        "description": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ]["properties"]["description"],
                                        "is_scheduled": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ]["properties"]["is_scheduled"],
                                        "schedule_window": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ][
                                            "properties"
                                        ][
                                            "schedule_window"
                                        ],
                                        "cron_schedule": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ]["properties"]["cron_schedule"],
                                        "dispatch.earliest_time": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ][
                                            "properties"
                                        ][
                                            "earliest_time"
                                        ],
                                        "dispatch.latest_time": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ][
                                            "properties"
                                        ][
                                            "latest_time"
                                        ],
                                    }

                                    ko_acl = {
                                        "owner": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ]["properties"]["eai:acl.owner"],
                                        "sharing": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ]["properties"]["eai:acl.sharing"],
                                        "perms.write": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ]["properties"]["eai:acl.perms.write"],
                                        "perms.read": savedsearches_restorable_dict[
                                            savedsearch_name
                                        ]["properties"]["eai:acl.perms.read"],
                                    }

                                    # Restore the savedsearch
                                    trackme_create_report(
                                        request_info.system_authtoken,
                                        request_info.server_rest_uri,
                                        tenant_id,
                                        savedsearch_name,
                                        savedsearches_restorable_dict[savedsearch_name][
                                            "definition"
                                        ],
                                        savedsearch_properties,
                                        ko_acl,
                                    )

                                    result = {
                                        "object": savedsearch_name,
                                        "object_type": "savedsearch",
                                        "action": "restore",
                                        "result": "success",
                                    }
                                    logging.info(json.dumps(result, indent=4))
                                    tasks_list.append(result)
                                    savedsearches_restored.append(
                                        savedsearch_name
                                    )  # Mark as restored

                                except Exception as e:
                                    result = {
                                        "object": savedsearch_name,
                                        "object_type": "savedsearch",
                                        "action": "restore",
                                        "result": "failure",
                                        "exception": str(e),
                                    }
                                    logging.error(json.dumps(result, indent=4))
                                    errors_list.append(result)

                        #############################################
                        # task: restore alerts definitions for tenant
                        #############################################

                        # Save available alerts in a main list
                        alerts_restorable_list_origin = []
                        for alert in alerts_restorable_dict:
                            alerts_restorable_list_origin.append(alert)

                        # If knowledge_objects_lists is not equal to "all", filter the list to its value
                        if knowledge_objects_lists != "all":
                            alerts_restorable_list = [
                                alert
                                for alert in alerts_restorable_list_origin
                                if alert in knowledge_objects_lists
                            ]
                        else:
                            alerts_restorable_list = (
                                alerts_restorable_list_origin.copy()
                            )

                        # Restoring saved searches has some challenges, if the search itself refers to another saved search, we need to
                        # restore this saved search first, and then restore the search that refers to it.

                        # Init a list of restored saved searches
                        alerts_restored = []

                        for alert_name in alerts_restorable_list:

                            logging.info(
                                f'TrackMe restore process starting, knowledge objects, processing restore of alert definition for alert_name="{alert_name}" for tenant_id="{tenant_id}"'
                            )

                            # First, get the search definition
                            alert_definition = alerts_restorable_dict[alert_name][
                                "definition"
                            ]

                            # Proceed with the restoration logic for the current saved search
                            alert_exists = False
                            try:
                                alert_object = service.saved_searches[alert_name]
                                alert_search = alert_object.content["search"]
                                alert_exists = True

                            except Exception:
                                pass

                            if alert_exists and not knowledge_objects_replace_existing:
                                logging.info(
                                    f'alert="{alert_name}" already exists and the replace_existing option is set to False, skipping the restore operation'
                                )
                                result = {
                                    "object": alert_name,
                                    "object_type": "alert",
                                    "action": "restore",
                                    "result": "skipped",
                                    "reason": "already exists and replace_existing is set to False",
                                }
                                logging.info(json.dumps(result, indent=4))
                                tasks_list.append(result)
                            else:
                                # Restore logic
                                try:
                                    if alert_exists:
                                        # Delete existing alert if required
                                        action = trackme_delete_report(
                                            request_info.system_authtoken,
                                            request_info.server_rest_uri,
                                            tenant_id,
                                            alert_name,
                                        )
                                        logging.info(
                                            f'alert="{alert_name}" was deleted successfully, response="{action}"'
                                        )

                                    # Prepare properties for restoration
                                    properties = {
                                        "description": alerts_restorable_dict[
                                            alert_name
                                        ]["properties"]["description"],
                                        "is_scheduled": alerts_restorable_dict[
                                            alert_name
                                        ]["properties"]["is_scheduled"],
                                        "schedule_window": alerts_restorable_dict[
                                            alert_name
                                        ]["properties"]["schedule_window"],
                                        "cron_schedule": alerts_restorable_dict[
                                            alert_name
                                        ]["properties"]["cron_schedule"],
                                        "dispatch.earliest_time": alerts_restorable_dict[
                                            alert_name
                                        ][
                                            "properties"
                                        ][
                                            "earliest_time"
                                        ],
                                        "dispatch.latest_time": alerts_restorable_dict[
                                            alert_name
                                        ]["properties"]["latest_time"],
                                    }

                                    ko_acl = {
                                        "owner": alerts_restorable_dict[alert_name][
                                            "properties"
                                        ]["eai:acl.owner"],
                                        "sharing": alerts_restorable_dict[alert_name][
                                            "properties"
                                        ]["eai:acl.sharing"],
                                        "perms.write": alerts_restorable_dict[
                                            alert_name
                                        ]["properties"]["eai:acl.perms.write"],
                                        "perms.read": alerts_restorable_dict[
                                            alert_name
                                        ]["properties"]["eai:acl.perms.read"],
                                    }

                                    alert_properties = alerts_restorable_dict[
                                        alert_name
                                    ]["alert_properties"]

                                    # Restore the alert
                                    trackme_create_alert(
                                        request_info.system_authtoken,
                                        request_info.server_rest_uri,
                                        tenant_id,
                                        alert_name,
                                        alerts_restorable_dict[alert_name][
                                            "definition"
                                        ],
                                        properties,
                                        alert_properties,
                                        ko_acl,
                                    )

                                    result = {
                                        "object": alert_name,
                                        "object_type": "alert",
                                        "action": "restore",
                                        "result": "success",
                                    }
                                    logging.info(json.dumps(result, indent=4))
                                    tasks_list.append(result)
                                    alerts_restored.append(
                                        alert_name
                                    )  # Mark as restored

                                except Exception as e:
                                    result = {
                                        "object": alert_name,
                                        "object_type": "alert",
                                        "action": "restore",
                                        "result": "failure",
                                        "exception": str(e),
                                    }
                                    logging.error(json.dumps(result, indent=4))
                                    errors_list.append(result)

                #######################################
                # Task: Restore the KVstore collections
                #######################################

                if (
                    restore_kvstore_collections
                    and kvstore_collections_restore_non_tenants_collections
                ):
                    logging.info(
                        f"TrackMe restore process starting, processing restore of KVstore collections"
                    )

                    # Loop through the restorable collections, and proceed
                    for collection_name in kvstore_collections_to_be_restored:

                        if (
                            collection_name not in kvstore_collections_restored
                            and collection_name != "kv_trackme_backup_archives_info"
                        ):  # restore collections that were not restored with the knowledge objects for the tenant
                            (
                                kvstore_collections_global_records_to_be_restored,
                                kvstore_collections_global_records_restored,
                                kvstore_collection_restore_summary_dict,
                            ) = restore_kvstore_records(
                                service,
                                collection_name,
                                collections_restore_dict,
                                backupdir,
                                kvstore_collections_global_records_to_be_restored,
                                kvstore_collections_global_records_restored,
                                kvstore_collections_restored_warning,
                                restore_results_dict,
                                kvstore_collections_clean_empty,
                            )

                            # add to dict
                            kvstore_collections_restore_summary_dict[
                                collection_name
                            ] = kvstore_collection_restore_summary_dict

                #######################################
                # End: clean up the temporary directory
                #######################################

                clean_backup_dir(backupdir)

            # investigate the results of the kvstore records restoration
            if (
                kvstore_collections_global_records_to_be_restored
                == kvstore_collections_global_records_restored
            ):
                kvstore_records_restore_summary = {
                    "response": "KVstore records restore operation is now complete and no errors were reported",
                    "action": "success",
                    "total_records_to_be_restored": kvstore_collections_global_records_to_be_restored,
                    "total_records_restored": kvstore_collections_global_records_restored,
                    "kvstore_collections_restore_summary_dict": kvstore_collections_restore_summary_dict,
                }

            else:
                kvstore_records_restore_summary = {
                    "response": "KVstore records restore operation is now complete but some errors were reported",
                    "action": "warning",
                    "message": "one or more KVstore collection could not be fully restored",
                    "total_records_to_be_restored": kvstore_collections_global_records_to_be_restored,
                    "total_records_restored": kvstore_collections_global_records_restored,
                    "kvstore_collections_restored_warning": kvstore_collections_restored_warning,
                    "kvstore_collections_restore_summary_dict": kvstore_collections_restore_summary_dict,
                }

            ######################
            # task: final response
            ######################

            # define action value, if errors_list is empty, set success, otherwise set warning
            action = "success" if not errors_list else "warning"
            response_dict["action"] = action

            # define response
            response = (
                "restore operation is now complete and no errors were reported, please reload TrackMe"
                if not errors_list
                else "restore operation is now complete but some errors were reported, please review these issues and reload TrackMe"
            )

            # add backupfile name
            response_dict["backupfile"] = backupfile

            # add the archive schema version
            response_dict["archive_schema_version"] = archive_schema_version

            # add the kvstore records restore results
            response_dict["kvstore_records_restore_summary"] = (
                kvstore_records_restore_summary
            )

            # add tasks list
            response_dict["tasks_executed"] = tasks_list

            # add errors list
            response_dict["tasks_errors"] = errors_list

            logging.info(json.dumps(response_dict, indent=4))
            return {"payload": response_dict, "status": 200}

    def post_import_backup(self, request_info, **kwargs):
        """
        Handles the import_backup functionality.
        """
        describe = False

        # Retrieve the payload
        try:
            resp_dict = json.loads(request_info.raw_args.get("payload", "{}"))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict.get("describe", False)
                if str(describe).lower() == "true":
                    describe = True
            except Exception:
                describe = False

        if describe:
            response = {
                "describe": "This endpoint imports a backup from a Base64-encoded TGZ file.",
                "resource_desc": "Import a backup archive",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/backup_and_restore/import_backup body=\"{'archive_base64': '<base64_string>'}\"",
                "options": [
                    {
                        "archive_base64": "(string) REQUIRED: Base64 string of the backup archive."
                    },
                ],
            }
            return {"payload": response, "status": 200}

        # Process the import
        try:
            archive_base64 = resp_dict.get("archive_base64")
            if not archive_base64:
                return {
                    "payload": {"error": "archive_base64 is required"},
                    "status": 400,
                }

            # Decode base64 and write to a temporary file
            decoded_data = base64.b64decode(archive_base64)
            temp_dir = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")
            os.makedirs(temp_dir, exist_ok=True)
            temp_file_path = os.path.join(temp_dir, "temp_import.tgz")

            with open(temp_file_path, "wb") as temp_file:
                temp_file.write(decoded_data)

            # Validate the archive
            if not tarfile.is_tarfile(temp_file_path):
                os.remove(temp_file_path)
                return {"payload": {"error": "Invalid TGZ archive"}, "status": 400}

            with tarfile.open(temp_file_path, "r:gz") as tar:
                members = tar.getnames()
                full_metadata_found = False
                light_metadata_found = False

                # check for metadata files
                for member in members:
                    if member.startswith("trackme-backup-") and member.endswith(
                        "full.meta"
                    ):
                        full_metadata_found = True
                        # extract the archive_name from the full metadata file, as: <filename including .tgz extension>.full.meta
                        archive_name = member.split(".full.meta")[0]
                        break

                for member in members:
                    if member.startswith("trackme-backup-") and member.endswith(
                        "light.meta"
                    ):
                        light_metadata_found = True
                        break

                if not full_metadata_found or not light_metadata_found:
                    os.remove(temp_file_path)
                    return {
                        "payload": {
                            "error": f"Missing metadata files, this archive is or corrupted, or not a TrackMe backup, or not created with TrackMe 2.1.5 and later (only archive_schema_version 2.0.0 and later can be imported), see available archive content: {members}"
                        },
                        "status": 400,
                    }

            # Extract the archive in backupdir
            try:
                backupdir = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")
                os.makedirs(backupdir, exist_ok=True)
                with tarfile.open(temp_file_path, "r:gz") as tar:
                    tar.extractall(backupdir)
                os.remove(temp_file_path)

            except Exception as e:
                os.remove(temp_file_path)
                return {
                    "payload": {
                        "error": f"Failed to extract the archive, error={str(e)}"
                    },
                    "status": 500,
                }

            # Call get_restore

            # url
            url = f"{request_info.server_rest_uri}/services/trackme/v2/backup_and_restore/backup"

            # header
            header = {
                "Authorization": "Splunk %s" % request_info.session_key,
                "Content-Type": "application/json",
            }

            response = requests.get(
                url,
                headers=header,
                verify=False,
                timeout=300,
            )

            if response.status_code != 200:
                return {
                    "payload": {
                        "error": f"Failed to call get_restore, http.status={response.status_code}, http.response={response.text}"
                    },
                    "status": 500,
                }

            else:
                return {
                    "payload": {"success": "Backup imported successfully"},
                    "status": 200,
                }

        except Exception as e:
            logger.error(f"Error in post_import_backup: {str(e)}")
            return {"payload": {"error": str(e)}, "status": 500}

    def post_export_backup(self, request_info, **kwargs):
        """
        Handles the export_backup functionality.
        """
        describe = False

        # Retrieve the payload
        try:
            resp_dict = json.loads(request_info.raw_args.get("payload", "{}"))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict.get("describe", False)
                if str(describe).lower() == "true":
                    describe = True
            except Exception:
                describe = False

        if describe:
            response = {
                "describe": "This endpoint exports a backup to a Base64-encoded TGZ file.",
                "resource_desc": "Export a backup archive",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/backup_and_restore/export_backup body="{\\"archive_name\\": \\"trackme-backup-20210205-142635.tgz\\"}"',
                "options": [
                    {
                        "archive_name": "(string) REQUIRED: Name of the backup archive to export."
                    },
                ],
            }
            return {"payload": response, "status": 200}

        # Process the export
        try:
            archive_name = resp_dict.get("archive_name")
            if not archive_name:
                return {"payload": {"error": "archive_name is required"}, "status": 400}

            # Locate the archive
            backup_dir = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")
            archive_path = os.path.join(backup_dir, archive_name)

            if not os.path.exists(archive_path):
                return {
                    "payload": {
                        "error": f"Archive not found, file {archive_path} does not exists."
                    },
                    "status": 404,
                }

            # Create a temporary TGZ archive
            temp_dir = os.path.join(splunkhome, "etc", "apps", "trackme", "backup")
            os.makedirs(temp_dir, exist_ok=True)
            temp_tgz_path = os.path.join(temp_dir, "export_archive.tgz")

            with tarfile.open(temp_tgz_path, "w:gz") as tar:
                tar.add(archive_path, arcname=archive_name)
                tar.add(
                    f"{archive_path}.light.meta", arcname=f"{archive_name}.light.meta"
                )
                tar.add(
                    f"{archive_path}.full.meta", arcname=f"{archive_name}.full.meta"
                )

            # Encode the archive in Base64
            with open(temp_tgz_path, "rb") as tgz_file:
                base64_data = base64.b64encode(tgz_file.read()).decode("utf-8")

            os.remove(temp_tgz_path)

            return {"payload": {"archive_base64": base64_data}, "status": 200}
        except Exception as e:
            logger.error(f"Error in post_export_backup: {str(e)}")
            return {"payload": {"error": str(e)}, "status": 500}
