#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_configuration.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import time
import requests
from collections import OrderedDict
from logging.handlers import RotatingFileHandler


# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import test handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import (
    trackme_create_report,
    trackme_delete_report,
    trackme_create_kvtransform,
    trackme_delete_kvtransform,
    trackme_create_macro,
    trackme_delete_macro,
    trackme_create_kvcollection,
    trackme_delete_kvcollection,
)

# import the collections dict
from collections_data import vtenant_account_default

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerConfiguration_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerConfiguration_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_configuration(self, request_info, **kwargs):
        response = {
            "resource_group_name": "configuration/admin",
            "resource_group_desc": "These endpoints provide various generic application level configuration capabilities (admin operations)",
        }

        return {"payload": response, "status": 200}

    # Create a Kvstore transforms with privileges escalation
    def post_create_kvtransform(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                transform_name = resp_dict["transform_name"]
                transform_fields = resp_dict["transform_fields"]
                collection_name = resp_dict["collection_name"]
                transform_acl = resp_dict["transform_acl"]
                owner = resp_dict["owner"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a KVstore transforms knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create KVstore transforms",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_kvstore_transforms" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"transform_name\\": \\"<transform_name>\\", \\"transform_fields\\": \\"<transform_fields>\\", \\"collection_name\\": \\"<collection_name>\\", \\"owner\\": \\"<owner>\\", \\"transform_acl\\": \\"<transform_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_kvtransform(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                transform_name,
                transform_fields,
                collection_name,
                owner,
                transform_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the transform definition, transform="{transform_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a Kvstore transforms with privileges escalation
    def post_delete_kvtransform(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                transform_name = resp_dict["transform_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a KVstore transforms knowledge object, it requires a POST with the following options:",
                "resource_desc": "Delete KVstore transforms",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_kvstore_transforms" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"transform_name\\": \\"<transform_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_kvtransform(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                transform_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the transform definition, transform="{transform_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a Kvstore collection with privileges escalation
    def post_delete_kvcollection(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                collection_name = resp_dict["collection_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a KVstore collection, it requires a POST with the following options:",
                "resource_desc": "Delete KVstore collection",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_kvstore_collection" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"collection_name\\": \\"<collection_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_kvcollection(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                collection_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the KVstore collection, collection="{collection_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Create a report with privileges escalation
    def post_create_report(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                report_name = resp_dict["report_name"]
                report_search = resp_dict["report_search"]
                report_properties = resp_dict["report_properties"]
                report_acl = resp_dict["report_acl"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a TrackMe report knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create TrackMe report",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_report" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"report_name\\": \\"<report_name>\\", \\"report_search\\": \\"<report_search>\\", \\"report_properties\\": \\"<report_properties>\\", \\"report_acl\\": \\"<report_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_report(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                report_name,
                report_search,
                report_properties,
                report_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the report definition, report="{report_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a report with privileges escalation
    def post_delete_report(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                report_name = resp_dict["report_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a TrackMe report knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create delete report",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_report" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"report_name\\": \\"<report_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_report(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                report_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the report definition, report="{report_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Update a report with privileges escalation
    def post_update_report(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                report_name = resp_dict["report_name"]
                report_search = resp_dict["report_search"]
                # Retrieving earliest and latest time from the request
                earliest_time = resp_dict.get("earliest_time")
                latest_time = resp_dict.get("latest_time")

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows updating a TrackMe report knowledge object, it requires a POST with the following options:",
                "resource_desc": "Update TrackMe report",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/update_report" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"report_name\\": \\"<report_name>\\", \\"report_search\\": \\"<report_search>\\"}"',
            }

            return {"payload": response, "status": 200}

        # get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=request_info.server_rest_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # update the report
        report_current = service.saved_searches[report_name]

        try:
            update_params = {"search": report_search}
            if earliest_time is not None:
                update_params["dispatch.earliest_time"] = earliest_time
            if latest_time is not None:
                update_params["dispatch.latest_time"] = latest_time

            action_update = report_current.update(**update_params)

            return {
                "payload": {
                    "action": "success",
                    "response": "The report was updated successfully",
                    "report_name": report_name,
                },
                "status": 200,
            }

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to update the report definition, report="{report_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Create a macro with privileges escalation
    def post_create_macro(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                macro_name = resp_dict["macro_name"]
                macro_definition = resp_dict["macro_definition"]
                macro_owner = resp_dict["macro_owner"]
                macro_acl = resp_dict["macro_acl"]
        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a macro knowledge object, it requires a POST with the following options:",
                "resource_desc": "Create macro",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_macro" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"macro_name\\": \\"<macro_name>\\", \\"macro_definition\\": \\"<macro_definition>\\", \\"macro_owner\\": \\"<macro_owner>\\", \\"macro_acl\\": \\"<macro_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_macro(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                macro_name,
                macro_definition,
                macro_owner,
                macro_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the macro definition, macro="{macro_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Update a macro with privileges escalation
    def post_update_macro(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                macro_name = resp_dict["macro_name"]
                macro_definition = resp_dict["macro_definition"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows updating a TrackMe macro knowledge object, it requires a POST with the following options:",
                "resource_desc": "update TrackMe macro",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/update_macro" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"macro_name\\": \\"<macro_name>\\", \\"macro_definition\\": \\"<macro_definition>\\"}"',
            }

            return {"payload": response, "status": 200}

        # get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=request_info.server_rest_port,
            token=request_info.system_authtoken,
            timeout=300,
        )

        # update the macro

        macro_current = service.confs["macros"][macro_name]

        try:
            action_update = macro_current.update(definition=macro_definition)
            return {
                "payload": {
                    "action": "success",
                    "response": "The macro was updated successfully",
                    "macro_name": macro_name,
                },
                "status": 200,
            }

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to update the macro definition, macro="{macro_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Delete a macro with privileges escalation
    def post_delete_macro(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                macro_name = resp_dict["macro_name"]

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows deleting a macro knowledge object, it requires a POST with the following options:",
                "resource_desc": "Delete macro",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/delete_macro" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"macro_name\\": \\"<macro_name>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_delete = trackme_delete_macro(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                macro_name,
            )
            return {"payload": action_delete, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to delete the macro definition, macro="{macro_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Create a KVstore collection with privileges escalation
    def post_create_kvcollection(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]
                collection_name = resp_dict["collection_name"]
                collection_acl = resp_dict["collection_acl"]
        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoint allows creating a Kvstore collection, it requires a POST with the following options:",
                "resource_desc": "Create KVstore collection",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/create_kvcollection" body="{\\"tenant_id\\": \\"<tenant_id>\\", \\"collection_name\\": \\"<collection_name>\\", \\"collection_acl\\": \\"<collection_acl>\\"}"',
            }

            return {"payload": response, "status": 200}

        # create the transform
        try:
            action_create = trackme_create_kvcollection(
                request_info.system_authtoken,
                request_info.server_rest_uri,
                tenant_id,
                collection_name,
                collection_acl,
            )
            return {"payload": action_create, "status": 200}

        except Exception as e:
            error_msg = f'tenant_id="{tenant_id}", failed to create the KVstore collection, collection="{collection_name}", exception="{str(e)}"'
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}

    # Verify and update the Virtual Tenant account with privileges escalation
    def post_maintain_vtenant_account(self, request_info, **kwargs):
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
                tenant_id = resp_dict["tenant_id"]

                # updated_vtenant_data is optional
                try:
                    updated_vtenant_data = resp_dict["updated_vtenant_data"]
                    # if not a dictionnary, attempt to load it
                    if not isinstance(updated_vtenant_data, dict):
                        try:
                            updated_vtenant_data = json.loads(updated_vtenant_data)
                        except Exception as e:
                            error_msg = f"failed to load updated_vtenant_data, exception={str(e)}"
                            logging.error(error_msg)
                            return {"payload": error_msg, "status": 500}

                except Exception as e:
                    updated_vtenant_data = None

                # force_create_missing is optional, accept true/false case insensitive and turn it into a boolean
                try:
                    force_create_missing = resp_dict["force_create_missing"]
                    if not isinstance(force_create_missing, bool):
                        force_create_missing = force_create_missing.lower()
                        if force_create_missing in ("true"):
                            force_create_missing = True
                        elif force_create_missing in ("false"):
                            force_create_missing = False
                        else:
                            return {
                                "payload": "force_create_missing must be a boolean, true or false",
                                "status": 500,
                            }

                except Exception as e:
                    force_create_missing = False

        else:
            describe = True

        # if describe is requested, show the usage
        if describe:
            response = {
                "describe": "This endpoints allows verifying and updating the Virtual Tenant account with privileges escalation, it requires a POST with the following options:",
                "resource_desc": "Create KVstore collection",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/maintain_vtenant_account" body="{\\"tenant_id\\": \\"<tenant_id>\\"}"',
                "options": [
                    {
                        "tenant_id": "The Virtual Tenant ID",
                        "updated_vtenant_data": "Optional, a dictionary with the updated Virtual Tenant data",
                        "force_create_missing": "Optional, if set to true, will force the creation of the Virtual Tenant account if it entirely missing, defaults to false.",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        #
        # main
        #

        url = f"{request_info.server_rest_uri}/servicesNS/nobody/trackme/trackme_vtenants/{tenant_id}"
        vtenant_data = {}

        # vtenant_account_found boolean
        vtenant_account_found = False

        # update boolean - if the current of the Virtual Tenants is missing any key from the default config, we need to update it
        update_is_required = False

        try:
            # Get current vtenant account configuration
            response = requests.get(
                url,
                headers={"Authorization": f"Splunk {request_info.system_authtoken}"},
                verify=False,
                params={"output_mode": "json"},
                timeout=300,
            )
            if response.status_code in (200, 201, 204):

                logging.info(f"successfully retrieved vtenant configuration")
                vtenant_data_json = response.json()
                vtenant_data_current = vtenant_data_json["entry"][0]["content"]

                # Set vtenant_account_found to True
                vtenant_account_found = True

                # Start with the current configuration and add missing keys from the default config
                # We keep all keys from the current configuration
                for key, value in vtenant_data_current.items():
                    if key in vtenant_account_default:
                        vtenant_data[key] = value

                # before merging with the default config, check if any key is missing
                for key in vtenant_account_default.keys():
                    if key not in vtenant_data:
                        update_is_required = True

                # Merge with default config, only adding missing default keys
                for key, value in vtenant_account_default.items():
                    if key not in vtenant_data:
                        vtenant_data[key] = value

                # If updated_vtenant_data is provided, it takes precedence over the defaults and current config
                if updated_vtenant_data:
                    vtenant_data.update(updated_vtenant_data)

                # Finally, ensures that each key in vtenant_data exists in vtenant_account_default, otherwise drop it
                vtenant_data = {
                    key: value
                    for key, value in vtenant_data.items()
                    if key in vtenant_account_default
                }

                logging.info(
                    f'vtenant_data="{json.dumps(vtenant_data, indent=2)}", update_is_required={update_is_required}'
                )

            else:
                error_msg = f"failed to retrieve vtenant configuration, status_code={response.status_code}"
                logging.error(error_msg)

        except Exception as e:
            error_msg = f"failed to retrieve vtenant configuration, exception={str(e)}"
            logging.error(error_msg)

        # init return_response
        return_response = {}

        #
        # if Virtual Tenant account is not found
        #

        # not found and force_create_missing is set to true
        if not vtenant_account_found and force_create_missing:
            url = f"{request_info.server_rest_uri}/servicesNS/nobody/trackme/trackme_vtenants"

            # load default vtenant config
            data = dict(vtenant_account_default)

            # add the name value
            data["name"] = tenant_id

            # Retrieve and set the tenant idx, if any failure, logs and use the global index
            try:
                response = requests.post(
                    url,
                    headers={
                        "Authorization": f"Splunk {request_info.system_authtoken}",
                        "Content-Type": "application/json",
                    },
                    data=data,
                    params={"output_mode": "json"},
                    verify=False,
                    timeout=300,
                )
                if response.status_code not in (200, 201, 204):

                    # set response
                    return_response["result"] = "failed"
                    return_response["message"] = (
                        f"failed to create vtenant account, status_code={response.status_code}, response={response.text}"
                    )
                    return_response["vtenant_account"] = None

                    # log
                    logging.error(return_response.get("message"))

                    # return return_response
                    return {"payload": return_response, "status": 500}

                else:

                    # set response
                    return_response["result"] = "success"
                    return_response["message"] = (
                        f"vtenant account created successfully, status_code={response.status_code}"
                    )
                    return_response["vtenant_account"] = data

                    # log
                    logging.info(return_response.get("message"))

                    # return response
                    return {"payload": return_response, "status": 200}

            except Exception as e:

                # set response
                return_response["result"] = "failed"
                return_response["message"] = (
                    f"failed to create vtenant account, exception={str(e)}"
                )
                return_response["vtenant_account"] = None

                # log
                logging.error(return_response.get("message"))

                # return response
                return {"payload": return_response, "status": 500}

        # not found and force_create_missing is not set to true
        elif not vtenant_account_found and not force_create_missing:

            # set response
            return_response["result"] = "failed"
            return_response["message"] = (
                f"vtenant account not found and force_create_missing is not set to true"
            )
            return_response["vtenant_account"] = None

            # log
            logging.error(return_response.get("message"))

            # return return_response
            return {"payload": return_response, "status": 500}

        #
        # main: Virtual Tenant account is found, if update is required, proceed with updating, otherwise return the account
        #

        # no update required
        if not update_is_required:

            # set response
            return_response["result"] = "success"
            return_response["message"] = (
                f"vtenant configuration checked successfully, no update required, status_code={response.status_code}"
            )
            return_response["vtenant_account"] = vtenant_data

            # log
            logging.info(return_response.get("message"))

            # return return_response
            return {"payload": return_response, "status": 200}

        # update required
        else:

            try:
                logging.info(
                    f'attempting to update vtenant configuration, vtenant_data="{json.dumps(vtenant_data, indent=2)}"'
                )
                response = requests.post(
                    url,
                    headers={
                        "Authorization": f"Splunk {request_info.system_authtoken}",
                        "Content-Type": "application/json",
                    },
                    data=vtenant_data,
                    verify=False,
                    timeout=300,
                )
                if response.status_code in (200, 201, 204):

                    # set response
                    return_response["result"] = "success"
                    return_response["message"] = (
                        f"vtenant configuration updated successfully, status_code={response.status_code}"
                    )
                    return_response["vtenant_account"] = vtenant_data

                    # log
                    logging.info(return_response.get("message"))

                    # return response
                    return {"payload": return_response, "status": 200}

                else:

                    # set return_response
                    return_response["result"] = "failed"
                    return_response["message"] = (
                        f"failed to update vtenant configuration, status_code={response.status_code}, response={response.text}"
                    )
                    return_response["vtenant_account"] = None

                    # log
                    logging.error(return_response.get("message"))

                    # return return_response
                    return {"payload": return_response, "status": 500}

            except Exception as e:

                # set response
                return_response["result"] = "failed"
                return_response["message"] = (
                    f"failed to update vtenant configuration, exception={str(e)}"
                )
                return_response["vtenant_account"] = None

                # log
                logging.error(return_response.get("message"))

                # return response
                return {"payload": return_response, "status": 500}

    # Retrieve the Virtual Tenant account configuration
    def post_get_vtenant_account(self, request_info, **kwargs):
        # Initialize response
        return_response = {}
        describe = False

        try:
            # Parse the request payload
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
            describe = resp_dict.get("describe", False)

            # Convert describe to boolean if it's a string
            if isinstance(describe, str) and describe.lower() in ("true", "false"):
                describe = describe.lower() == "true"

            tenant_id = resp_dict.get("tenant_id")
            if not tenant_id and not describe:
                return {
                    "payload": "Missing tenant_id in the request payload.",
                    "status": 400,
                }
        except Exception as e:
            if not describe:
                error_msg = f"Invalid payload format: {str(e)}"
                logging.error(error_msg)
                return {"payload": error_msg, "status": 400}

        # If describe is requested, return the endpoint usage details
        if describe:
            response = {
                "describe": "This endpoint retrieves the Virtual Tenant account configuration, it requires a POST with the following options",
                "resource_desc": "Retrieve Virtual Tenant account",
                "resource_spl_example": '| trackme mode=post url="/services/trackme/v2/configuration/admin/get_vtenant_account" body="{\\"tenant_id\\": \\"<tenant_id>\\"}"',
                "options": [
                    {
                        "tenant_id": "The Virtual Tenant ID",
                        "describe": "Optional, if set to true, returns the endpoint description and usage details.",
                    }
                ],
            }
            return {"payload": response, "status": 200}

        # Build the URL for retrieving the Virtual Tenant account
        url = f"{request_info.server_rest_uri}/servicesNS/nobody/trackme/trackme_vtenants/{tenant_id}"

        try:
            # Send a GET request to retrieve the account configuration
            response = requests.get(
                url,
                headers={"Authorization": f"Splunk {request_info.system_authtoken}"},
                verify=False,
                params={"output_mode": "json"},
                timeout=300,
            )

            if response.status_code in (200, 201, 204):
                # Parse the response JSON
                vtenant_data_json = response.json()

                # Extract the account content
                if "entry" in vtenant_data_json and len(vtenant_data_json["entry"]) > 0:
                    vtenant_data = vtenant_data_json["entry"][0]["content"]

                    # Set successful response
                    return_response["result"] = "success"
                    return_response["message"] = (
                        "Virtual Tenant account retrieved successfully."
                    )
                    return_response["vtenant_account"] = vtenant_data

                    logging.info(return_response["message"])
                    return {"payload": return_response, "status": 200}
                else:
                    # Handle case where account content is not found
                    return_response["result"] = "failed"
                    return_response["message"] = (
                        "Virtual Tenant account content not found."
                    )
                    return_response["vtenant_account"] = None

                    logging.error(return_response["message"])
                    return {"payload": return_response, "status": 404}
            else:
                # Handle HTTP error responses
                error_msg = f"Failed to retrieve Virtual Tenant account, status_code={response.status_code}, response={response.text}"
                logging.error(error_msg)
                return {"payload": error_msg, "status": response.status_code}

        except Exception as e:
            # Handle exceptions during the GET request
            error_msg = (
                f"Exception occurred while retrieving Virtual Tenant account: {str(e)}"
            )
            logging.error(error_msg)
            return {"payload": error_msg, "status": 500}
