#!/usr/bin/env python
# coding=utf-8

__name__ = "trackme_rest_handler_priority_policies.py"
__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Built-in libraries
import json
import logging
import os
import sys
import re
import time
from logging.handlers import RotatingFileHandler

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
logger = logging.getLogger(__name__)
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_rest_api.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()
for hdlr in log.handlers[:]:
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import rest handler
import trackme_rest_handler

# import trackme libs
from trackme_libs import trackme_getloglevel, trackme_audit_event, get_kv_collection

# import Splunk libs
import splunklib.client as client


class TrackMeHandlerSplkPriorityPolicies_v2(trackme_rest_handler.RESTHandler):
    def __init__(self, command_line, command_arg):
        super(TrackMeHandlerSplkPriorityPolicies_v2, self).__init__(
            command_line, command_arg, logger
        )

    def get_resource_group_desc_splk_priority_policies(self, request_info, **kwargs):
        response = {
            "resource_group_name": "splk_priority_policies/write",
            "resource_group_desc": "Endpoints related to the management of priorities (power operations)",
        }

        return {"payload": response, "status": 200}

    # Add new policy
    def post_priority_policies_add(self, request_info, **kwargs):

        # Declare
        tenant_id = None
        component = None
        priority_policy_id = None
        priority_policy_value = None
        priority_policy_regex = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:

                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The tenant_id is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm", "mhm", "wlk", "flx", "cim"):
                        return {
                            "payload": {
                                "action": "failure",
                                "response": f"invalid component {component}, valid options are: dsm/dhm/mhm/wlk/flx/cim",
                            },
                            "status": 400,
                        }
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The component is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    priority_policy_id = resp_dict["priority_policy_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The priority_policy_id is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    priority_policy_value = resp_dict["priority_policy_value"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The priority_policy_value is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                # if priority_policy_value is a string, turn into a list and make it lower case, if is a list, return an error as it is not accepted in this context
                if isinstance(priority_policy_value, str):
                    priority_policy_value = priority_policy_value.lower()
                    # check value, valid options: low, medium, high, critical
                    if priority_policy_value not in (
                        "low",
                        "medium",
                        "high",
                        "critical",
                    ):
                        return {
                            "payload": {
                                "response": "The priority_policy_value is not a valid value, valid options are: low, medium, high, critical",
                            },
                            "status": 400,
                        }
                else:
                    return {
                        "payload": {
                            "response": "The priority_policy_value should be a string",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    priority_policy_regex = resp_dict["priority_policy_regex"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The priority_policy_regex is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                # verify that the regex is valid
                try:
                    re.compile(priority_policy_regex)
                except re.error:
                    return {
                        "payload": {
                            "response": "The priority_policy_regex is not a valid regular expression",
                            "status": 400,
                        },
                        "status": 400,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint creates a new priority policy or updates a policy if it exists already, it requires a POST call with the following data:",
                "resource_desc": "Add or update a priority policy",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_priority_policies/write/priority_policies_add\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'priority_policy_id': 'linux_secure', 'priority_policy_id': 'linux_secure', 'priority_policy_value': 'Linux,OS,CIM', 'priority_policy_regex': '\:linux_secure$'}\"",
                "options": [
                    {
                        "tenant_id": "(required) The tenant identifier",
                        "component": "The component identifier, valid values are: dsm/dhm/mhm/wlk/flx/cim",
                        "priority_policy_id": "(required) ID of the priority policy",
                        "priority_policy_regex": "(required) The regular expression to be used by the priority policy, special characters should be escaped.",
                        "priority_policy_value": "(required) priority value (low/medium/high/critical) to be applied.",
                        "update_comment": "(optional) Comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Define the KV query
        query_string = {
            "priority_policy_id": priority_policy_id,
        }

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # Data collection
        collection_name = f"kv_trackme_{component}_priority_policies_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # Get the current record
        # Notes: the record is returned as an array, as we search for a specific record, we expect one record only

        try:
            record = collection.data.query(query=json.dumps(query_string))
            key = record[0].get("_key")

        except Exception as e:
            key = None

        # proceed

        # this policy exists already, it will be updated to include any missing priority from the list
        if key:
            # set an action_desc
            action_desc = "updated"

            # create a new list
            priority_new = []

            # add existing members
            for priority in record[0].get("priority_policy_value"):
                priority_new.append(priority)

            # add priority, if not in there already
            for priority in priority_policy_value:
                if priority not in priority_new:
                    priority_new.append(priority)

        else:
            # set an action_desc
            action_desc = "created"

        # proceed
        try:
            if key:
                # Update the record
                collection.data.update(
                    str(key),
                    json.dumps(
                        {
                            "priority_policy_id": priority_policy_id,
                            "priority_policy_value": priority_policy_value,
                            "priority_policy_regex": priority_policy_regex,
                            "mtime": time.time(),
                        }
                    ),
                )

                # Record an audit change
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "update priority policy",
                        str(priority_policy_id),
                        f"splk-{component}",
                        collection.data.query(query=json.dumps(query_string)),
                        "The priority policy was updated successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

            else:
                # Insert the record
                collection.data.insert(
                    json.dumps(
                        {
                            "priority_policy_id": priority_policy_id,
                            "priority_policy_value": priority_policy_value,
                            "priority_policy_regex": priority_policy_regex,
                            "mtime": time.time(),
                        }
                    )
                )

                # Record an audit change
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "success",
                        "add priority policy",
                        str(priority_policy_id),
                        f"splk-{component}",
                        collection.data.query(query=json.dumps(query_string)),
                        "The priority policy was added successfully",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

            # render response
            response = {
                "action": "success",
                "action_desc": action_desc,
                "response": 'the priority policy priority_policy_id="{}" was {} successfully'.format(
                    priority_policy_id, action_desc
                ),
                "record": collection.data.query(query=json.dumps(query_string)),
            }

            return {"payload": response, "status": 200}

        except Exception as e:
            response = {
                "action": "failure",
                "response": 'an exception was encountered, exception="{}"'.format(
                    str(e)
                ),
            }
            logging.error(json.dumps(response))
            return {"payload": response, "status": 500}

    # Delete records from the collection
    def post_priority_policies_del(self, request_info, **kwargs):
        # Declare
        tenant_id = None
        component = None
        priority_policy_id_list = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:

                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The tenant_id is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm", "mhm", "wlk", "flx", "cim"):
                        return {
                            "payload": {
                                "action": "failure",
                                "response": f"invalid component {component}, valid options are: dsm/dhm/mhm/wlk/flx/cim",
                            },
                            "status": 400,
                        }
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The component is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    priority_policy_id_list = resp_dict["priority_policy_id_list"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The priority_policy_id_list is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                # if priority_policy_id_list is a string, turn into a list
                if isinstance(priority_policy_id_list, str):
                    priority_policy_id_list = priority_policy_id_list.split(",")

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint deletes priority policies, it requires a POST call with the following information:",
                "resource_desc": "Delete or ore more priority policies",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_priority_policies/write/priority_policies_del\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'priority_policy_id': 'linux_secure,linux_sec'}\"",
                "options": [
                    {
                        "tenant_id": "(required) The tenant identifier",
                        "component": "The component identifier, valid values are: dsm/dhm/mhm/wlk/flx/cim",
                        "priority_policy_id_list": "(required) Comma separated list of priority policies",
                        "update_comment": "(optional) Comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        collection_name = f"kv_trackme_{component}_priority_policies_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # entities KV collection
        priority_collection_name = f"kv_trackme_{component}_priority_tenant_{tenant_id}"
        priority_collection = service.kvstore[priority_collection_name]

        # records summary
        records = []

        # loop
        for item in priority_policy_id_list:

            # Define the KV query
            query_string = {
                "priority_policy_id": item,
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # Remove and audit
                try:
                    # Remove the record
                    collection.data.delete(json.dumps({"_key": key}))

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "delete priority policy",
                            str(item),
                            "all",
                            record,
                            "The lagging class was deleted successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "success",
                        "record": record,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                    #
                    # clean up the priority collection
                    #

                    # Define the KV query
                    query_string = {
                        "priority_reason": item,
                    }

                    # Get records
                    priority_records = priority_collection.data.query(
                        query=json.dumps(query_string)
                    )

                    for priority_record in priority_records:
                        priority_record_key = priority_record.get("_key")

                        # silently remove the record
                        try:
                            priority_collection.data.delete(
                                json.dumps({"_key": priority_record_key})
                            )
                        except:
                            pass

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "delete priority policy",
                            str(item),
                            "all",
                            record,
                            str(e),
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "failure",
                        "record": record,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "delete priority policy",
                        str(item),
                        "all",
                        record,
                        "HTTP 404 NOT FOUND",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

                result = {
                    "action": "delete",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Update records
    def post_priority_policies_update(self, request_info, **kwargs):
        # Declare
        tenant_id = None
        component = None
        records_list = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The tenant_id is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm", "mhm", "wlk", "flx", "cim"):
                        return {
                            "payload": {
                                "action": "failure",
                                "response": f"invalid component {component}, valid options are: dsm/dhm/mhm/wlk/flx/cim",
                            },
                            "status": 400,
                        }
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The component is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    records_list = resp_dict["records_list"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The records_list is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    records_list = json.loads(records_list)
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The records_list is not a valid JSON",
                            "status": 400,
                        },
                        "status": 400,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint updates records, it requires a POST call with the following information:",
                "resource_desc": "Update or ore more priority policies",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_priority_policies/write/priority_policies_update\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm', 'records_list': '<redacted_json_records>'}\"",
                "options": [
                    {
                        "tenant_id": "(required) The tenant identifier",
                        "component": "The component identifier, valid values are: dsm/dhm/mhm/wlk/flx/cim",
                        "records_list": "(required) JSON records to be updated",
                        "update_comment": "(option) Comment for the update, comments are added to the audit record, if unset will be defined to: API update",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Update comment is optional and used for audit changes
        try:
            update_comment = resp_dict["update_comment"]
        except Exception as e:
            update_comment = "API update"

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # counters
        processed_count = 0
        succcess_count = 0
        failures_count = 0

        # Data collection
        collection_name = f"kv_trackme_{component}_priority_policies_tenant_{tenant_id}"
        collection = service.kvstore[collection_name]

        # records summary
        records = []

        # debug
        logging.info('records_list="{}"'.format(json.dumps(records_list, indent=0)))

        # loop
        for item in records_list:
            # debug
            logging.info('item="{}"'.format(item))

            # Define the KV query
            query_string = {
                "priority_policy_id": item.get("priority_policy_id"),
            }

            # Get the current record
            # Notes: the record is returned as an array, as we search for a specific record, we expect one record only
            try:
                record = collection.data.query(query=json.dumps(query_string))
                key = record[0].get("_key")

            except Exception as e:
                key = None

            # Render result
            if key:
                # This record exists already

                # Store the record for audit purposes
                record = str(json.dumps(collection.data.query_by_id(key), indent=1))

                # Update and audit
                try:
                    # Update the record
                    priority_policy_value = item.get("priority_policy_value")

                    # if priority_policy_value is a string, make it lower case, if is a list, return an error as it is not accepted in this context
                    if isinstance(priority_policy_value, str):
                        priority_policy_value = priority_policy_value.lower()
                        # check value, valid options: low, medium, high, critical
                        if priority_policy_value not in (
                            "low",
                            "medium",
                            "high",
                            "critical",
                        ):
                            return {
                                "payload": {
                                    "response": "The priority_policy_value is not a valid value, valid options are: low, medium, high, critical",
                                },
                                "status": 400,
                            }
                    else:
                        return {
                            "payload": {
                                "response": "The priority_policy_value should be a string",
                                "status": 400,
                            },
                            "status": 400,
                        }

                    collection.data.update(
                        str(key),
                        json.dumps(
                            {
                                "priority_policy_id": item.get("priority_policy_id"),
                                "priority_policy_value": priority_policy_value,
                                "priority_policy_regex": item.get(
                                    "priority_policy_regex"
                                ),
                                "mtime": time.time(),
                            }
                        ),
                    )

                    # increment counter
                    processed_count += 1
                    succcess_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "success",
                            "update priority policies",
                            str(item),
                            "dsm",
                            record,
                            "The lagging classs was updated successfully",
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "success",
                        "record": record,
                    }

                    records.append(result)

                    logging.info(json.dumps(result, indent=0))

                except Exception as e:
                    # increment counter
                    processed_count += 1
                    succcess_count += 0
                    failures_count += 1

                    # audit record
                    try:
                        trackme_audit_event(
                            request_info.system_authtoken,
                            request_info.server_rest_uri,
                            tenant_id,
                            request_info.user,
                            "failure",
                            "update priority policies",
                            str(item),
                            "dsm",
                            record,
                            str(e),
                            str(update_comment),
                        )
                    except Exception as e:
                        logging.error(
                            'failed to generate an audit event with exception="{}"'.format(
                                str(e)
                            )
                        )

                    result = {
                        "action": "delete",
                        "result": "failure",
                        "record": record,
                        "exception": e,
                    }

                    # append to records
                    records.append(result)

                    # log
                    logging.error(json.dumps(result, indent=0))

            else:
                # increment counter
                processed_count += 1
                succcess_count += 0
                failures_count += 1

                # audit record
                try:
                    trackme_audit_event(
                        request_info.system_authtoken,
                        request_info.server_rest_uri,
                        tenant_id,
                        request_info.user,
                        "failure",
                        "update priority policies",
                        str(item),
                        "dsm",
                        record,
                        "HTTP 404 NOT FOUND",
                        str(update_comment),
                    )
                except Exception as e:
                    logging.error(
                        'failed to generate an audit event with exception="{}"'.format(
                            str(e)
                        )
                    )

                result = {
                    "action": "delete",
                    "result": "failure",
                    "record": item,
                    "exception": "HTTP 404 NOT FOUND",
                }

                # append to records
                records.append(result)

                # log
                logging.error(json.dumps(result, indent=0))

        # render HTTP status and summary

        req_summary = {
            "process_count": processed_count,
            "success_count": succcess_count,
            "failures_count": failures_count,
            "records": records,
        }

        if processed_count > 0 and processed_count == succcess_count:
            return {"payload": req_summary, "status": 200}

        else:
            return {"payload": req_summary, "status": 500}

    # Simulate priority
    def post_priority_policies_simulate(self, request_info, **kwargs):
        # Declare
        tenant_id = None
        component = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The tenant_id is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm", "mhm", "wlk", "flx", "cim"):
                        return {
                            "payload": {
                                "action": "failure",
                                "response": f"invalid component {component}, valid options are: dsm/dhm/mhm/wlk/flx/cim",
                            },
                            "status": 400,
                        }
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The component is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    regex_value = resp_dict["regex_value"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The argument regex_value is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    priority = resp_dict["priority"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The argument priority is required",
                            "status": 400,
                        },
                        "status": 400,
                    }
                # priority should be a string, valid options: low, medium, high, critical
                if isinstance(priority, str):
                    priority = priority.lower()
                    # check value, valid options: low, medium, high, critical
                    if priority not in ("low", "medium", "high", "critical"):
                        return {
                            "payload": {
                                "response": "The priority is not a valid value, valid options are: low, medium, high, critical",
                            },
                            "status": 400,
                        }
                else:
                    return {
                        "payload": {
                            "response": "The priority should be a string",
                            "status": 400,
                        },
                        "status": 400,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint simulates a priority policy, it requires a POST call with the following information:",
                "resource_desc": "Simulates a priority policy",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_priority_policies/write/priority_policies_simulate\" body=\"{'tenant_id': 'mytenant', 'regex_value': '^org_eu.*', 'priority': 'high'}\"",
                "options": [
                    {
                        "tenant_id": "(required) The tenant identifier",
                        "regex_value": "(required) The regex to be used for the simulation",
                        "priority": "(required) The priority value (low/medium/high/critical)",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # start
        main_start = time.time()

        #
        # priority policies
        #

        priority_policies_records = [
            {
                "priority_policy_regex": regex_value,
                "priority_policy_value": priority,
            }
        ]

        #
        # KV entities
        #

        # entities KV collection
        data_collection_name = f"kv_trackme_{component}_tenant_{tenant_id}"
        data_collection = service.kvstore[data_collection_name]

        # get records
        data_records, data_collection_keys, data_collection_dict = get_kv_collection(
            data_collection, data_collection_name
        )

        #
        # Handle policies
        #

        # A policy is composed by the following fields: priority_policy_id, priority_policy_regex, priority_policy_value
        # A record has various fields, in this context we care about: object (the value against the regex is applied) and priority_auto (a list to contain all matching priority values)

        # Counters and error list
        entities_failures_count = 0
        entities_exceptions_list = []
        entities_matched = []

        # only proceed if we have entities
        if len(data_records) > 0:

            # loop through records and apply policies
            for entity_record in data_records:

                entity_key = entity_record["_key"]
                entity_object = entity_record["object"]

                # apply policies
                if len(priority_policies_records) > 0:
                    for policy_record in priority_policies_records:

                        # check that the regex expression is a valid regex expression
                        # for the purpose of the simulation, we will however return a 200 with a message
                        try:
                            re.compile(regex_value)
                        except re.error:
                            req_summary = {
                                "entities_matched_count": len(entities_matched),
                                "result_summary": "The regex_value is not a valid regular expression, review this expression and try again.",
                                "regex_value": regex_value,
                                "regex_is_valid": "false",
                            }
                            return {"payload": req_summary, "status": 200}

                        try:
                            # apply regex
                            regex = policy_record["priority_policy_regex"]
                            value = policy_record["priority_policy_value"]

                            # make it lower case
                            if isinstance(value, str):
                                value = value.lower()

                            if re.match(regex, entity_record["object"]):
                                logging.info(
                                    f'tenant_id="{tenant_id}", object="{entity_object}", regex has matched this entity, regex="{regex}", value="{value}"'
                                )

                                # append to our list
                                if not entity_object in entities_matched:
                                    entities_matched.append(entity_object)

                            else:
                                logging.info(
                                    f'tenant_id="{tenant_id}", object="{entity_object}", regex has not matched this entity, regex="{regex}", value="{value}"'
                                )

                        except Exception as e:
                            logging.error(
                                f'context="exception", failed to apply policy, exception="{str(e)}"'
                            )

        # set action
        if entities_failures_count == 0:
            action = "success"
        else:
            action = "failure"

        # get run_time
        run_time = round((time.time() - main_start), 3)

        # create a result summary
        if len(entities_matched) > 0:
            result_summary = f"The regex has matched {len(entities_matched)} entities."
        else:
            result_summary = "The regex has not matched any entities, verify your inputs and try again."

        # request summary
        req_summary = {
            "kvstore_collection_entities_count": len(data_records),
            "entities_matched_count": len(entities_matched),
            "entities_matched": entities_matched,
            "result_summary": result_summary,
            "error_messages": entities_exceptions_list,
            "regex_is_valid": "true",
        }

        # render response
        if action == "success":
            logging.info(
                f'priority simulation operation has terminated, action="{action}", tenant_id="{tenant_id}", run_time="{run_time}"'
            )
            return {"payload": req_summary, "status": 200}

        else:
            logging.error(
                f'priority simulation operation has failed, action="{action}", tenant_id="{tenant_id}", req_summary="{json.dumps(req_summary, indent=2)}"'
            )
            return {"payload": req_summary, "status": 500}

    # Apply priority
    def post_priority_policies_apply(self, request_info, **kwargs):
        # Declare
        tenant_id = None
        component = None
        describe = False

        # Retrieve from data
        try:
            resp_dict = json.loads(str(request_info.raw_args["payload"]))
        except Exception as e:
            resp_dict = None

        if resp_dict is not None:
            try:
                describe = resp_dict["describe"]
                if describe in ("true", "True"):
                    describe = True
            except Exception as e:
                describe = False
            if not describe:
                try:
                    tenant_id = resp_dict["tenant_id"]
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The tenant_id is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

                try:
                    component = resp_dict["component"]
                    if component not in ("dsm", "dhm", "mhm", "wlk", "flx", "cim"):
                        return {
                            "payload": {
                                "action": "failure",
                                "response": f"invalid component {component}, valid options are: dsm/dhm/mhm/wlk/flx/cim",
                            },
                            "status": 400,
                        }
                except Exception as e:
                    return {
                        "payload": {
                            "response": "The component is required",
                            "status": 400,
                        },
                        "status": 400,
                    }

        else:
            # body is required in this endpoint, if not submitted describe the usage
            describe = True

        if describe:
            response = {
                "describe": "This endpoint applies priority policies, it requires a POST call with the following information:",
                "resource_desc": "Immediately apply and update priority policies",
                "resource_spl_example": "| trackme mode=post url=\"/services/trackme/v2/splk_priority_policies/write/priority_policies_apply\" body=\"{'tenant_id': 'mytenant', 'component': 'dsm'}\"",
                "options": [
                    {
                        "tenant_id": "(required) The tenant identifier",
                        "component": "The component identifier, valid values are: dsm/dhm/mhm/wlk/flx/cim",
                    }
                ],
            }

            return {"payload": response, "status": 200}

        # Get splunkd port
        splunkd_port = request_info.server_rest_port

        # Get service
        service = client.connect(
            owner="nobody",
            app="trackme",
            port=splunkd_port,
            token=request_info.session_key,
            timeout=300,
        )

        # set loglevel
        loglevel = trackme_getloglevel(
            request_info.system_authtoken, request_info.server_rest_port
        )
        log.setLevel(logging.getLevelName(loglevel))

        # start
        main_start = time.time()

        #
        # KV priority policies
        #

        # priority policies KV collection
        priority_policies_collection_name = (
            f"kv_trackme_{component}_priority_policies_tenant_{tenant_id}"
        )
        priority_policies_collection = service.kvstore[
            priority_policies_collection_name
        ]

        # get records
        (
            priority_policies_records,
            priority_collection_keys,
            priority_collection_dict,
        ) = get_kv_collection(
            priority_policies_collection, priority_policies_collection_name
        )

        #
        # KV entities
        #

        # entities KV collection
        data_collection_name = f"kv_trackme_{component}_tenant_{tenant_id}"
        data_collection = service.kvstore[data_collection_name]

        # get records
        data_records, data_collection_keys, data_collection_dict = get_kv_collection(
            data_collection, data_collection_name
        )

        #
        # KV priority
        #

        # entities KV collection
        priority_collection_name = f"kv_trackme_{component}_priority_tenant_{tenant_id}"
        priority_collection = service.kvstore[priority_collection_name]

        # get records
        priority_records, priority_collection_keys, priority_collection_dict = (
            get_kv_collection(priority_collection, priority_collection_name)
        )

        #
        # Handle policies
        #

        # A policy is composed by the following fields: priority_policy_id, priority_policy_regex, priority_policy_value
        # A record has various fields, in this context we care about: object (the value against the regex is applied) and priority_auto (a list to contain all matching priority values)

        # Counters and error list
        entities_updated_count = 0
        entities_failures_count = 0
        entities_deleted_count = 0
        entities_exceptions_list = []

        updated_records = []  # we will store the updated records here

        # dict representing the numerical conversion of the priority value (low: 1, medium: 2, high: 3, critical: 4)
        priority_dict = {"low": 1, "medium": 2, "high": 3, "critical": 4}

        # only proceed if we have entities
        if len(data_records) > 0:

            # loop through records and apply policies
            for entity_record in data_records:

                entity_key = entity_record["_key"]
                entity_object = entity_record["object"]
                priority = None
                priority_reason = None

                # apply policies
                if len(priority_policies_records) > 0:

                    # Initialize priority and reason as None
                    priority = None
                    priority_reason = None
                    policy_matched = False

                    for policy_record in priority_policies_records:
                        try:
                            regex = policy_record["priority_policy_regex"]
                            value = policy_record["priority_policy_value"].lower()

                            if re.match(regex, entity_record["object"]):
                                logging.info(
                                    f'tenant_id="{tenant_id}", object="{entity_object}", policy="{policy_record["priority_policy_id"]}" has matched this entity, regex="{regex}", value="{value}"'
                                )

                                # Check if a priority was already matched
                                if policy_matched:
                                    # Get the numerical value of the matched priority
                                    priority_matched_num = priority_dict.get(
                                        priority, -1
                                    )  # Get -1 if none set

                                    # Get the numerical value of the current priority
                                    priority_current_num = priority_dict.get(value, -1)

                                    # Update if the new priority value is higher
                                    if priority_current_num > priority_matched_num:
                                        priority = value
                                        priority_reason = policy_record[
                                            "priority_policy_id"
                                        ]
                                else:
                                    priority = value
                                    priority_reason = policy_record[
                                        "priority_policy_id"
                                    ]
                                    policy_matched = True

                        except Exception as e:
                            logging.error(
                                f'context="exception", failed to apply policy, exception="{str(e)}"'
                            )

                # Add if matched
                if policy_matched:

                    # add to updated_records
                    updated_records.append(
                        {
                            "_key": entity_key,
                            "object": entity_object,
                            "priority": priority,
                            "priority_reason": priority_reason,
                            "mtime": time.time(),
                        }
                    )

            # Update records in batches
            chunks = [
                updated_records[i : i + 500]
                for i in range(0, len(updated_records), 500)
            ]
            for chunk in chunks:
                try:
                    priority_collection.data.batch_save(*chunk)
                    entities_updated_count += len(chunk)
                except Exception as e:
                    msg = f'KVstore batch save failed with exception="{str(e)}"'
                    logging.error(msg)
                    entities_exceptions_list.append(msg)

            # for record in priority_records, if the key of the record does not exist in data_collection_keys, delete it
            for record in priority_records:
                if record["_key"] not in data_collection_keys:
                    try:
                        priority_collection.data.delete(
                            json.dumps({"_key": record["_key"]})
                        )
                        # counter deleted
                        entities_deleted_count += 1

                    except Exception as e:
                        # counter failure
                        entities_failures_count += 1
                        msg = f'KVstore delete failed for key="{record["_key"]}", exception="{str(e)}"'
                        logging.error(msg)
                        entities_exceptions_list.append(msg)

            # refresh knowledge
            priority_records, priority_collection_keys, priority_collection_dict = (
                get_kv_collection(priority_collection, priority_collection_name)
            )

            # immediately apply priority on the data collection, loop through priority_records and apply priority to data_records
            updated_data_records = []  # we will store the updated records here

            for priority_record in priority_records:
                for entity_record in data_records:
                    if entity_record["_key"] == priority_record["_key"]:

                        # get the current priority value, we will only update the KVstore record if the final priority field has changed
                        current_priority = entity_record.get("priority", None)

                        entity_record["priority"] = priority_record["priority"]

                        # compare current_priority (CSV string) and priority (CSV string), if they are different, add the record to the updated_data_records list
                        if current_priority:
                            if current_priority != priority_record["priority"]:
                                entity_record["mtime"] = time.time()
                                updated_data_records.append(entity_record)

                        else:  # we have no current priority, we need to update the record
                            entity_record["mtime"] = time.time()
                            updated_data_records.append(entity_record)

        # set action
        if entities_failures_count == 0:
            action = "success"
        else:
            action = "failure"

        # get run_time
        run_time = round((time.time() - main_start), 3)

        # request summary
        req_summary = {
            "tenant_id": tenant_id,
            "action": action,
            "run_time": run_time,
            "kvstore_lookup_collection": f"trackme_{component}_priority_tenant_{tenant_id}",
            "priority_policies_no_records": len(priority_policies_records),
            "kvstore_collection_entities_count": len(data_records),
            "entities_updated_count": entities_updated_count,
            "entities_failures_count": entities_failures_count,
            "entities_deleted_count": entities_deleted_count,
            "error_messages": entities_exceptions_list,
        }

        # render response
        if action == "success":
            logging.info(
                f'priority apply operation has terminated, action="{action}", tenant_id="{tenant_id}", run_time="{run_time}"'
            )
            return {"payload": req_summary, "status": 200}

        else:
            logging.error(
                f'priority apply operation has failed, action="{action}", tenant_id="{tenant_id}", req_summary="{json.dumps(req_summary, indent=2)}"'
            )
            return {"payload": req_summary, "status": 500}
