#!/usr/bin/env python
# coding=utf-8

__author__ = "TrackMe Limited"
__copyright__ = "Copyright 2023-2024, TrackMe Limited, U.K."
__credits__ = "TrackMe Limited, U.K."
__license__ = "TrackMe Limited, all rights reserved"
__version__ = "0.1.0"
__maintainer__ = "TrackMe Limited, U.K."
__email__ = "support@trackme-solutions.com"
__status__ = "PRODUCTION"

# Standard library
import os
import sys
import json
import time

# Logging
import logging
from logging.handlers import RotatingFileHandler

# Networking
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# splunk home
splunkhome = os.environ["SPLUNK_HOME"]

# set logging
filehandler = RotatingFileHandler(
    "%s/var/log/splunk/trackme_decision_maker.log" % splunkhome,
    mode="a",
    maxBytes=10000000,
    backupCount=1,
)
formatter = logging.Formatter(
    "%(asctime)s %(levelname)s %(filename)s %(funcName)s %(lineno)d %(message)s"
)
logging.Formatter.converter = time.gmtime
filehandler.setFormatter(formatter)
log = logging.getLogger()  # root logger - Good to get it only once.
for hdlr in log.handlers[:]:  # remove the existing file handlers
    if isinstance(hdlr, logging.FileHandler):
        log.removeHandler(hdlr)
log.addHandler(filehandler)  # set the new handler
# set the log level to INFO, DEBUG as the default is ERROR
log.setLevel(logging.INFO)

# append lib
sys.path.append(os.path.join(splunkhome, "etc", "apps", "trackme", "lib"))

# import Splunk
from splunklib.searchcommands import (
    dispatch,
    StreamingCommand,
    Configuration,
    Option,
    validators,
)

# Import trackme libs
from trackme_libs import trackme_reqinfo, trackme_vtenant_account

# import TrackMe get data libs
from trackme_libs_get_data import (
    get_feeds_datagen_kv_collection,
    get_full_kv_collection,
    get_sampling_kv_collection,
)

# Import trackme decisionmaker libs
from trackme_libs_decisionmaker import (
    convert_epoch_to_datetime,
    get_monitoring_wdays,
    get_outliers_status,
    get_data_sampling_status,
    get_future_status,
    get_future_metrics_status,
    get_is_under_dcount_host,
    get_logical_groups_collection_records,
    get_dsm_latency_status,
    get_dsm_delay_status,
    set_dsm_status,
    set_dhm_status,
    set_mhm_status,
    set_flx_status,
    set_wlk_status,
    set_cim_status,
    apply_blocklist,
    dynamic_priority_lookup,
    dynamic_tags_lookup,
    dynamic_sla_class_lookup,
    get_sla_timer,
    dsm_sampling_lookup,
    sampling_anomaly_status,
)


@Configuration(distributed=False)
class TrackMeDecisionMaker(StreamingCommand):
    tenant_id = Option(
        doc="""
        **Syntax:** **tenant_id=****
        **Description:** The tenant identifier.""",
        require=True,
        default=None,
    )

    component = Option(
        doc="""
        **Syntax:** **component=****
        **Description:** Specify the TrackMe component.""",
        require=True,
        default=None,
        validate=validators.Match("component", r"^(dsm|dhm|mhm|wlk|flx|cim)$"),
    )

    """
    This function ensures that records have the same list of fields to allow Splunk to automatically extract these fields
    If a given result does not have a given field, it will be added to the record as an empty value    
    """

    def generate_fields(self, records):
        all_keys = set()
        for record in records:
            all_keys.update(record.keys())

        for record in records:
            for key in all_keys:
                if key not in record:
                    record[key] = ""
            yield record

    """
    Stream function
    """

    def stream(self, records):
        # Start performance counter
        start = time.time()

        # Get request info and set logging level
        reqinfo = trackme_reqinfo(
            self._metadata.searchinfo.session_key, self._metadata.searchinfo.splunkd_uri
        )
        log.setLevel(reqinfo["logging_level"])

        # Get virtual tenant account
        vtenant_conf = trackme_vtenant_account(
            self._metadata.searchinfo.session_key,
            self._metadata.searchinfo.splunkd_uri,
            self.tenant_id,
        )

        #
        # System level settings
        #

        system_future_tolerance = float(
            reqinfo["trackme_conf"]["splk_general"][
                "splk_general_feeds_future_tolerance"
            ]
        )

        # dynamic priority, for all components
        # get priority collection
        priority_collection_name = (
            f"kv_trackme_{self.component}_priority_tenant_{self.tenant_id}"
        )
        priority_collection = self.service.kvstore[priority_collection_name]
        priority_records, priority_collection_keys, priority_collection_dict = (
            get_full_kv_collection(priority_collection, priority_collection_name)
        )

        # get tags collection
        tags_collection_name = (
            f"kv_trackme_{self.component}_tags_tenant_{self.tenant_id}"
        )
        tags_collection = self.service.kvstore[tags_collection_name]
        tags_records, tags_collection_keys, tags_collection_dict = (
            get_full_kv_collection(tags_collection, tags_collection_name)
        )

        # dynamic sla_class, for all components
        # get sla collection
        sla_collection_name = f"kv_trackme_{self.component}_sla_tenant_{self.tenant_id}"
        sla_collection = self.service.kvstore[sla_collection_name]
        sla_records, sla_collection_keys, sla_collection_dict = get_full_kv_collection(
            sla_collection, sla_collection_name
        )

        #
        # SLA timer
        #

        sla_classes = {}
        sla_default_class = None

        sla_classes = reqinfo["trackme_conf"]["sla"]["sla_classes"]
        # try loading the JSON
        try:
            sla_classes = json.loads(sla_classes)
            sla_default_class = reqinfo["trackme_conf"]["sla"]["sla_default_class"]
            if not len(sla_default_class) > 0 or sla_default_class not in sla_classes:
                sla_default_class = "silver"
                logging.error(
                    f'Invalid sla_default_class="{sla_default_class}", this SLA class is not part of the SLA classes, applying fallback configuration'
                )

        except:
            logging.error(
                f'Error loading sla_classes JSON, please check the configuration, the JSON is not valid JSON, applying fallback configuration, exception="{str(e)}"'
            )
            sla_classes = json.loads(
                '{"gold": {"sla_threshold": 14400, "rank": 3}, "silver": {"sla_threshold": 86400, "rank": 2}, "platinum": {"sla_threshold": 172800, "rank": 1}}'
            )
            sla_default_class = "silver"

        #
        # splk-dsm specific collections
        #

        if self.component == "dsm":

            # Data sampling
            sampling_collection_name = (
                f"kv_trackme_dsm_data_sampling_tenant_{self.tenant_id}"
            )
            sampling_collection = self.service.kvstore[sampling_collection_name]
            sampling_records, sampling_collection_keys, sampling_collection_dict = (
                get_sampling_kv_collection(
                    sampling_collection, sampling_collection_name
                )
            )

        # dhm specific

        if self.component == "dhm":
            macro_name = (
                f"trackme_dhm_default_splk_dhm_alert_policy_tenant_{self.tenant_id}"
            )
            macro_current = self.service.confs["macros"][macro_name]
            default_splk_dhm_alerting_policy = macro_current.content.get("definition")
            # remove double quotes from default_splk_dhm_alerting_policy
            default_splk_dhm_alerting_policy = default_splk_dhm_alerting_policy.replace(
                '"', ""
            )

            logging.debug(
                f'default_splk_dhm_alerting_policy="{default_splk_dhm_alerting_policy}"'
            )

        #
        # feeds specific collections (splk-dsm/splk-dhm/splk-mhm)
        #

        if self.component in ["dsm", "dhm", "mhm", "flx", "wlk"]:

            # datagen
            datagen_collection_name = (
                f"kv_trackme_{self.component}_allowlist_tenant_{self.tenant_id}"
            )
            datagen_collection = self.service.kvstore[datagen_collection_name]
            (
                datagen_records,
                datagen_collection_keys,
                datagen_collection_dict,
                datagen_collection_blocklist_not_regex_dict,
                datagen_collection_blocklist_regex_dict,
            ) = get_feeds_datagen_kv_collection(
                datagen_collection, datagen_collection_name, self.component
            )

            logging.debug(
                f'datagen_collection_dict="{json.dumps(datagen_collection_dict, indent=2)}"'
            )

            logging.debug(
                f'datagen_collection_blocklist_not_regex_dict="{json.dumps(datagen_collection_blocklist_not_regex_dict, indent=2)}"'
            )

            logging.debug(
                f'datagen_collection_blocklist_regex_dict="{json.dumps(datagen_collection_blocklist_regex_dict, indent=2)}"'
            )

        #
        # Virtual tenant account settings
        #

        # outliers tenant level settings
        tenant_outliers_set_state = int(vtenant_conf["outliers_set_state"])
        tenant_data_sampling_set_state = int(vtenant_conf["data_sampling_set_state"])

        #
        # Logical groups collection records
        #

        logical_group_coll = self.service.kvstore[
            f"kv_trackme_common_logical_group_tenant_{self.tenant_id}"
        ]

        (
            logical_coll_records,
            logical_coll_dict,
            logical_coll_members_list,
            logical_coll_members_dict,
            logical_coll_count,
        ) = get_logical_groups_collection_records(logical_group_coll)

        # log debug
        logging.debug(
            f'function get_logical_groups_collection_records, logical_coll_dict="{json.dumps(logical_coll_dict, indent=2)}", logical_coll_count="{logical_coll_count}"'
        )

        # Process records
        processed_records = []
        records_count = 0
        logging.debug(f"processing records")
        for record in records:
            records_count += 1
            try:
                new_record = {}

                # append_record boolean, True by default unless specific use cases
                append_record = True

                logging.debug(f"processing record")

                # get object_value and key
                object_value = record.get("object", None)
                logging.debug(
                    f'object="{object_value}", record="{json.dumps(record, indent=2)}"'
                )

                # save the current value of object_state in the record as kvcurrent_object_state, we manipulate real state calculations
                # and we need the original state in some conditions (sla)
                record["kvcurrent_object_state"] = record.get("object_state", "N/A")

                # The value for key is normally in the field keyid, but in some cases it is in the field key or _key
                # use keyid, key, _key in that order of preference
                key_value = record.get("keyid", None)
                if key_value is None:
                    key_value = record.get("key", None)
                if key_value is None:
                    key_value = record.get("_key", None)

                #
                # Dynamic priority
                #

                dynamic_priority_lookup(
                    key_value,
                    priority_collection_keys,
                    priority_collection_dict,
                    record,
                )

                #
                # Dynamic tags
                #

                dynamic_tags_lookup(
                    key_value,
                    tags_collection_keys,
                    tags_collection_dict,
                    record,
                )

                #
                # Dynamic sla_class
                #

                dynamic_sla_class_lookup(
                    key_value,
                    sla_collection_keys,
                    sla_collection_dict,
                    record,
                )

                #
                # splk-dsm
                #

                # get record fields depending on the component
                if self.component == "dsm":

                    # first check blocklist
                    if (
                        datagen_collection_blocklist_not_regex_dict
                        or datagen_collection_blocklist_regex_dict
                    ):
                        append_record = apply_blocklist(
                            record,
                            datagen_collection_blocklist_not_regex_dict,
                            datagen_collection_blocklist_regex_dict,
                        )

                    if append_record:

                        # get outliers and data sampling
                        try:
                            isOutlier = int(record.get("isOutlier", 0))
                        except:
                            isOutlier = 0

                        try:
                            OutliersDisabled = int(record.get("OutliersDisabled", 0))
                        except:
                            OutliersDisabled = 0

                        try:
                            isAnomaly = int(record.get("isAnomaly", 0))
                        except:
                            isAnomaly = 0

                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", isAnomaly="{isAnomaly}"'
                        )

                        # get future_tolerance
                        future_tolerance = record.get("future_tolerance", 0)
                        try:
                            future_tolerance = float(future_tolerance)
                        except:
                            future_tolerance = 0

                        #
                        # DSM Sampling
                        #

                        # call function dsm_sampling_lookup
                        dsm_sampling_lookup(
                            object_value,
                            sampling_collection_keys,
                            sampling_collection_dict,
                            record,
                        )

                        # get data_lag_alert_kpis
                        data_lag_alert_kpis = record.get(
                            "data_lag_alert_kpis", "all_kpis"
                        )

                        # get actual primary KPI values
                        data_last_ingestion_lag_seen = record.get(
                            "data_last_ingestion_lag_seen", 0
                        )
                        data_last_lag_seen = record.get("data_last_lag_seen", 0)

                        # get per entity thresholds
                        data_max_lag_allowed = float(
                            record.get("data_max_lag_allowed", 0)
                        )
                        data_max_delay_allowed = float(
                            record.get("data_max_delay_allowed", 0)
                        )
                        min_dcount_threshold = record.get("min_dcount_threshold", 0)
                        try:
                            min_dcount_threshold = float(min_dcount_threshold)
                        except:
                            min_dcount_threshold = 0

                        # get dcount host related information
                        min_dcount_host = record.get("min_dcount_host", "any")
                        try:
                            min_dcount_host = float(min_dcount_host)
                        except:
                            pass
                        min_dcount_field = record.get("min_dcount_field", None)

                        # get monitoring wdays and hours ranges
                        data_monitoring_wdays = record.get(
                            "data_monitoring_wdays", "auto:all_days"
                        )
                        data_monitoring_hours_ranges = record.get(
                            "data_monitoring_hours_ranges", "auto:all_ranges"
                        )

                        # Get logical group information

                        # get logical group information: object_group_key
                        object_group_key = record.get("object_group_key", "")

                        # from logical_coll_dict, get object_logical_group_dict by object_group_key, this is sent to the status function
                        object_logical_group_dict = logical_coll_dict.get(
                            object_group_key, {}
                        )

                        # get data_last_ingest, data_last_time_seen, data_last_time_seen_idx (epochtime)
                        data_last_ingest = record.get("data_last_ingest", 0)
                        try:
                            data_last_ingest = float(data_last_ingest)
                        except:
                            pass
                        data_last_time_seen = record.get("data_last_time_seen", 0)
                        try:
                            data_last_time_seen = float(data_last_time_seen)
                        except:
                            pass
                        data_last_time_seen_idx = record.get(
                            "data_last_time_seen_idx", 0
                        )
                        try:
                            data_last_time_seen_idx = float(data_last_time_seen_idx)
                        except:
                            pass

                        # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                        (
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                        ) = get_monitoring_wdays(
                            data_monitoring_wdays, data_monitoring_hours_ranges
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", data_monitoring_wdays="{data_monitoring_wdays}", data_monitoring_hours_ranges="{data_monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                        )

                        # call get_outliers_status and define isOutlier
                        isOutlier = get_outliers_status(
                            isOutlier, OutliersDisabled, tenant_outliers_set_state
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", OutliersDisabled="{OutliersDisabled}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                        )

                        # call get_data_sampling_status and define isAnomaly
                        isAnomaly = get_data_sampling_status(
                            isAnomaly, tenant_data_sampling_set_state
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isAnomaly="{isAnomaly}", tenant_data_sampling_set_state="{tenant_data_sampling_set_state}"'
                        )

                        # call get_future_status and define isFuture
                        (
                            isFuture,
                            isFutureMsg,
                            merged_future_tolerance,
                        ) = get_future_status(
                            future_tolerance,
                            system_future_tolerance,
                            data_last_lag_seen,
                            data_last_ingestion_lag_seen,
                            data_last_time_seen,
                            data_last_ingest,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isFuture="{isFuture}", future_tolerance="{future_tolerance}", system_future_tolerance="{system_future_tolerance}", merged_future_tolerance="{merged_future_tolerance}", data_last_lag_seen="{data_last_lag_seen}", isFutureMsg="{isFutureMsg}"'
                        )

                        # call get_is_under_dcount_host and define isUnderDcountHost
                        (
                            isUnderDcountHost,
                            isUnderDcountHostMsg,
                        ) = get_is_under_dcount_host(
                            min_dcount_host, min_dcount_threshold, min_dcount_field
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderDcountHost="{isUnderDcountHost}", isUnderDcountHostMsg="{isUnderDcountHostMsg}", min_dcount_host="{min_dcount_host}", min_dcount_threshold="{min_dcount_threshold}"'
                        )

                        # call get_dsm_latency_status and define isUnderLatencyAlert and isUnderLatencyMessage
                        (
                            isUnderLatencyAlert,
                            isUnderLatencyMessage,
                        ) = get_dsm_latency_status(
                            data_last_ingestion_lag_seen,
                            data_max_lag_allowed,
                            data_lag_alert_kpis,
                            data_last_ingest,
                            data_last_time_seen,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLatencyAlert="{isUnderLatencyAlert}", isUnderLatencyMessage="{isUnderLatencyMessage}", data_last_ingestion_lag_seen="{data_last_ingestion_lag_seen}", data_max_lag_allowed="{data_max_lag_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                        )

                        # call get_dsm_delay_status and define isUnderDelayAlert and isUnderDelayMessage
                        (
                            isUnderDelayAlert,
                            isUnderDelayMessage,
                        ) = get_dsm_delay_status(
                            data_last_lag_seen,
                            data_max_delay_allowed,
                            data_lag_alert_kpis,
                            data_last_ingest,
                            data_last_time_seen,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderDelayAlert="{isUnderDelayAlert}", isUnderDelayMessage="{isUnderDelayMessage}", data_last_lag_seen="{data_last_lag_seen}", data_max_delay_allowed="{data_max_delay_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                        )

                        # call set_dsm_status and define object_state and anomaly_reason
                        (
                            object_state,
                            status_message,
                            status_message_json,
                            anomaly_reason,
                        ) = set_dsm_status(
                            self._metadata.searchinfo.splunkd_uri,
                            self._metadata.searchinfo.session_key,
                            self.tenant_id,
                            record,
                            isOutlier,
                            isAnomaly,
                            isFuture,
                            isFutureMsg,
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                            isUnderDcountHost,
                            isUnderDcountHostMsg,
                            object_logical_group_dict,
                            isUnderLatencyAlert,
                            isUnderLatencyMessage,
                            isUnderDelayAlert,
                            isUnderDelayMessage,
                        )
                        logging.debug(
                            f'set_dsm_status, tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                        )

                        # insert our main fields
                        new_record["object_state"] = object_state
                        new_record["status_message"] = " | ".join(status_message)
                        new_record["status_message_json"] = status_message_json
                        new_record["anomaly_reason"] = "|".join(anomaly_reason)

                        # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                        if isUnderMonitoringDays:
                            new_record["isUnderMonitoringDays"] = "True"
                        else:
                            new_record["isUnderMonitoringDays"] = "False"
                        if isUnderMonitoringHours:
                            new_record["isUnderMonitoringHours"] = "True"
                        else:
                            new_record["isUnderMonitoringHours"] = "False"

                        # future tolerance
                        try:
                            new_record["future_tolerance"] = int(
                                round(merged_future_tolerance, 0)
                            )
                        except:
                            new_record["future_tolerance"] = -600

                        # convert data_last_time_seen to last_time from epoch
                        last_time = convert_epoch_to_datetime(data_last_time_seen)
                        new_record["last_time"] = last_time

                        # convert data_last_ingest to last_ingest from epoch
                        last_ingest = convert_epoch_to_datetime(data_last_ingest)
                        new_record["last_ingest"] = last_ingest

                        # convert data_last_time_seen_idx to last_time_idx from epoch
                        last_time_idx = convert_epoch_to_datetime(data_last_time_seen)
                        new_record["last_time_idx"] = last_time_idx

                        # get and convert latest_flip_time from epoch
                        latest_flip_time_human = record.get("latest_flip_time", 0)
                        try:
                            latest_flip_time_human = float(latest_flip_time_human)
                        except:
                            latest_flip_time_human = 0
                        new_record["latest_flip_time_human"] = (
                            convert_epoch_to_datetime(latest_flip_time_human)
                        )

                        # sla_timer
                        get_sla_timer(record, sla_classes, sla_default_class)

                #
                # splk-dhm
                #

                elif self.component == "dhm":

                    # first check blocklist
                    if (
                        datagen_collection_blocklist_not_regex_dict
                        or datagen_collection_blocklist_regex_dict
                    ):
                        append_record = apply_blocklist(
                            record,
                            datagen_collection_blocklist_not_regex_dict,
                            datagen_collection_blocklist_regex_dict,
                        )

                    if append_record:

                        # get splk_dhm_st_summary
                        splk_dhm_st_summary = record.get("splk_dhm_st_summary", None)
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", splk_dhm_st_summary="{splk_dhm_st_summary}"'
                        )

                        # get outliers and data sampling
                        try:
                            isOutlier = int(record.get("isOutlier", 0))
                        except:
                            isOutlier = 0

                        try:
                            OutliersDisabled = int(record.get("OutliersDisabled", 0))
                        except:
                            OutliersDisabled = 0

                        try:
                            isAnomaly = int(record.get("isAnomaly", 0))
                        except:
                            isAnomaly = 0

                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", isAnomaly="{isAnomaly}"'
                        )

                        # get future_tolerance
                        future_tolerance = record.get("future_tolerance", 0)
                        try:
                            future_tolerance = float(future_tolerance)
                        except:
                            future_tolerance = 0

                        # get data_lag_alert_kpis
                        data_lag_alert_kpis = record.get(
                            "data_lag_alert_kpis", "all_kpis"
                        )

                        # get actual primary KPI values
                        data_last_ingestion_lag_seen = record.get(
                            "data_last_ingestion_lag_seen", 0
                        )
                        data_last_lag_seen = record.get("data_last_lag_seen", 0)

                        # get per entity thresholds
                        data_max_lag_allowed = float(
                            record.get("data_max_lag_allowed", 0)
                        )
                        data_max_delay_allowed = float(
                            record.get("data_max_delay_allowed", 0)
                        )

                        # get monitoring wdays and hours ranges
                        data_monitoring_wdays = record.get(
                            "data_monitoring_wdays", "auto:all_days"
                        )
                        data_monitoring_hours_ranges = record.get(
                            "data_monitoring_hours_ranges", "auto:all_ranges"
                        )

                        # Get logical group information

                        # get logical group information: object_group_key
                        object_group_key = record.get("object_group_key", "")

                        # from logical_coll_dict, get object_logical_group_dict by object_group_key, this is sent to the status function
                        object_logical_group_dict = logical_coll_dict.get(
                            object_group_key, {}
                        )

                        # get data_last_ingest, data_last_time_seen, data_last_time_seen_idx (epochtime)
                        data_last_ingest = record.get("data_last_ingest", 0)
                        try:
                            data_last_ingest = float(data_last_ingest)
                        except:
                            pass
                        data_last_time_seen = record.get("data_last_time_seen", 0)
                        try:
                            data_last_time_seen = float(data_last_time_seen)
                        except:
                            pass
                        data_last_time_seen_idx = record.get(
                            "data_last_time_seen_idx", 0
                        )
                        try:
                            data_last_time_seen_idx = float(data_last_time_seen_idx)
                        except:
                            pass

                        # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                        (
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                        ) = get_monitoring_wdays(
                            data_monitoring_wdays, data_monitoring_hours_ranges
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", data_monitoring_wdays="{data_monitoring_wdays}", data_monitoring_hours_ranges="{data_monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                        )

                        # call get_outliers_status and define isOutlier
                        isOutlier = get_outliers_status(
                            isOutlier, OutliersDisabled, tenant_outliers_set_state
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", OutliersDisabled="{OutliersDisabled}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                        )

                        # call get_future_status and define isFuture
                        (
                            isFuture,
                            isFutureMsg,
                            merged_future_tolerance,
                        ) = get_future_status(
                            future_tolerance,
                            system_future_tolerance,
                            data_last_lag_seen,
                            data_last_ingestion_lag_seen,
                            data_last_time_seen,
                            data_last_ingest,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isFuture="{isFuture}", future_tolerance="{future_tolerance}", system_future_tolerance="{system_future_tolerance}", merged_future_tolerance="{merged_future_tolerance}", data_last_lag_seen="{data_last_lag_seen}", isFutureMsg="{isFutureMsg}"'
                        )

                        # call get_dsm_latency_status and define isUnderLatencyAlert and isUnderLatencyMessage
                        (
                            isUnderLatencyAlert,
                            isUnderLatencyMessage,
                        ) = get_dsm_latency_status(
                            data_last_ingestion_lag_seen,
                            data_max_lag_allowed,
                            data_lag_alert_kpis,
                            data_last_ingest,
                            data_last_time_seen,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderLatencyAlert="{isUnderLatencyAlert}", isUnderLatencyMessage="{isUnderLatencyMessage}", data_last_ingestion_lag_seen="{data_last_ingestion_lag_seen}", data_max_lag_allowed="{data_max_lag_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                        )

                        # call get_dsm_delay_status and define isUnderDelayAlert and isUnderDelayMessage
                        (
                            isUnderDelayAlert,
                            isUnderDelayMessage,
                        ) = get_dsm_delay_status(
                            data_last_lag_seen,
                            data_max_delay_allowed,
                            data_lag_alert_kpis,
                            data_last_ingest,
                            data_last_time_seen,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isUnderDelayAlert="{isUnderDelayAlert}", isUnderDelayMessage="{isUnderDelayMessage}", data_last_lag_seen="{data_last_lag_seen}", data_max_delay_allowed="{data_max_delay_allowed}", data_lag_alert_kpis="{data_lag_alert_kpis}", data_last_ingest="{data_last_ingest}", data_last_time_seen="{data_last_time_seen}"'
                        )

                        # call set_dhm_status and define object_state and anomaly_reason
                        (
                            object_state,
                            status_message,
                            status_message_json,
                            anomaly_reason,
                            splk_dhm_alerting_policy,
                        ) = set_dhm_status(
                            self._metadata.searchinfo.splunkd_uri,
                            self._metadata.searchinfo.session_key,
                            self.tenant_id,
                            record,
                            isOutlier,
                            isFuture,
                            isFutureMsg,
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                            object_logical_group_dict,
                            isUnderLatencyAlert,
                            isUnderLatencyMessage,
                            isUnderDelayAlert,
                            isUnderDelayMessage,
                            default_splk_dhm_alerting_policy,
                            data_lag_alert_kpis,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                        )

                        # insert our main fields
                        new_record["object_state"] = object_state
                        new_record["status_message"] = " | ".join(status_message)
                        new_record["status_message_json"] = status_message_json
                        new_record["anomaly_reason"] = "|".join(anomaly_reason)

                        # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                        if isUnderMonitoringDays:
                            new_record["isUnderMonitoringDays"] = "True"
                        else:
                            new_record["isUnderMonitoringDays"] = "False"
                        if isUnderMonitoringHours:
                            new_record["isUnderMonitoringHours"] = "True"
                        else:
                            new_record["isUnderMonitoringHours"] = "False"

                        # future tolerance
                        try:
                            new_record["future_tolerance"] = int(
                                round(merged_future_tolerance, 0)
                            )
                        except:
                            new_record["future_tolerance"] = -600

                        # specific for dhm
                        new_record["splk_dhm_alerting_policy"] = (
                            splk_dhm_alerting_policy
                        )

                        # convert data_last_time_seen to last_time from epoch
                        last_time = convert_epoch_to_datetime(data_last_time_seen)
                        new_record["last_time"] = last_time

                        # convert data_last_ingest to last_ingest from epoch
                        last_ingest = convert_epoch_to_datetime(data_last_ingest)
                        new_record["last_ingest"] = last_ingest

                        # convert data_last_time_seen_idx to last_time_idx from epoch
                        last_time_idx = convert_epoch_to_datetime(data_last_time_seen)
                        new_record["last_time_idx"] = last_time_idx

                        # get and convert latest_flip_time from epoch
                        latest_flip_time_human = record.get("latest_flip_time", 0)
                        try:
                            latest_flip_time_human = float(latest_flip_time_human)
                        except:
                            latest_flip_time_human = 0
                        new_record["latest_flip_time_human"] = (
                            convert_epoch_to_datetime(latest_flip_time_human)
                        )

                        # sla_timer
                        get_sla_timer(record, sla_classes, sla_default_class)

                #
                # splk-mhm
                #

                elif self.component == "mhm":

                    # first check blocklist
                    if (
                        datagen_collection_blocklist_not_regex_dict
                        or datagen_collection_blocklist_regex_dict
                    ):
                        append_record = apply_blocklist(
                            record,
                            datagen_collection_blocklist_not_regex_dict,
                            datagen_collection_blocklist_regex_dict,
                        )

                    if append_record:

                        # get metric_details
                        metric_details = record.get("metric_details", None)
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", metric_details="{metric_details}"'
                        )

                        # Get logical group information

                        # get logical group information: object_group_key
                        object_group_key = record.get("object_group_key", "")

                        # from logical_coll_dict, get object_logical_group_dict by object_group_key, this is sent to the status function
                        object_logical_group_dict = logical_coll_dict.get(
                            object_group_key, {}
                        )

                        # get metric_last_time_seen (epochtime)
                        metric_last_time_seen = record.get("metric_last_time_seen", 0)
                        try:
                            metric_last_time_seen = float(metric_last_time_seen)
                        except:
                            pass

                        # call get_future_metrics_status and define isFuture
                        isFuture, isFutureMsg = get_future_metrics_status(
                            system_future_tolerance,
                            metric_last_time_seen,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isFuture="{isFuture}", system_future_tolerance="{system_future_tolerance}", metric_last_time_seen="{metric_last_time_seen}", isFutureMsg="{isFutureMsg}"'
                        )

                        # call set_mhm_status and define object_state and anomaly_reason
                        (
                            object_state,
                            status_message,
                            status_message_json,
                            anomaly_reason,
                        ) = set_mhm_status(
                            self._metadata.searchinfo.splunkd_uri,
                            self._metadata.searchinfo.session_key,
                            self.tenant_id,
                            record,
                            metric_details,
                            isFuture,
                            isFutureMsg,
                            object_logical_group_dict,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                        )

                        # insert our main fields
                        new_record["object_state"] = object_state
                        new_record["status_message"] = " | ".join(status_message)
                        new_record["status_message_json"] = status_message_json
                        new_record["anomaly_reason"] = "|".join(anomaly_reason)

                        # convert metric_last_time_seen to last_time from epoch
                        last_time = convert_epoch_to_datetime(metric_last_time_seen)
                        new_record["last_time"] = last_time

                        # get and convert latest_flip_time from epoch
                        latest_flip_time_human = record.get("latest_flip_time", 0)
                        try:
                            latest_flip_time_human = float(latest_flip_time_human)
                        except:
                            latest_flip_time_human = 0
                        new_record["latest_flip_time_human"] = (
                            convert_epoch_to_datetime(latest_flip_time_human)
                        )

                        # sla_timer
                        get_sla_timer(record, sla_classes, sla_default_class)

                #
                # splk-flx
                #

                # get record fields depending on the component
                elif self.component == "flx":

                    # first check blocklist
                    if (
                        datagen_collection_blocklist_not_regex_dict
                        or datagen_collection_blocklist_regex_dict
                    ):
                        append_record = apply_blocklist(
                            record,
                            datagen_collection_blocklist_not_regex_dict,
                            datagen_collection_blocklist_regex_dict,
                        )

                    if append_record:

                        # get outliers
                        try:
                            isOutlier = int(record.get("isOutlier", 0))
                        except:
                            isOutlier = 0

                        try:
                            OutliersDisabled = int(record.get("OutliersDisabled", 0))
                        except:
                            OutliersDisabled = 0

                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}"'
                        )

                        # get monitoring wdays and hours ranges
                        monitoring_wdays = record.get(
                            "monitoring_wdays", "auto:all_days"
                        )
                        monitoring_hours_ranges = record.get(
                            "monitoring_hours_ranges", "auto:all_ranges"
                        )

                        # Get logical group information

                        # get logical group information: object_group_key
                        object_group_key = record.get("object_group_key", "")

                        # from logical_coll_dict, get object_logical_group_dict by object_group_key, this is sent to the status function
                        object_logical_group_dict = logical_coll_dict.get(
                            object_group_key, {}
                        )

                        # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                        (
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                        ) = get_monitoring_wdays(
                            monitoring_wdays, monitoring_hours_ranges
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", monitoring_wdays="{monitoring_wdays}", monitoring_hours_ranges="{monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                        )

                        # call get_outliers_status and define isOutlier
                        isOutlier = get_outliers_status(
                            isOutlier, OutliersDisabled, tenant_outliers_set_state
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", OutliersDisabled="{OutliersDisabled}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                        )

                        # call set_flx_status and define object_state and anomaly_reason
                        (
                            object_state,
                            status_message,
                            status_message_json,
                            anomaly_reason,
                        ) = set_flx_status(
                            self._metadata.searchinfo.splunkd_uri,
                            self._metadata.searchinfo.session_key,
                            self.tenant_id,
                            record,
                            isOutlier,
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                            object_logical_group_dict,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                        )

                        # insert our main fields
                        new_record["object_state"] = object_state
                        new_record["status_message"] = " | ".join(status_message)
                        new_record["status_message_json"] = status_message_json
                        new_record["anomaly_reason"] = "|".join(anomaly_reason)

                        # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                        if isUnderMonitoringDays:
                            new_record["isUnderMonitoringDays"] = "True"
                        else:
                            new_record["isUnderMonitoringDays"] = "False"
                        if isUnderMonitoringHours:
                            new_record["isUnderMonitoringHours"] = "True"
                        else:
                            new_record["isUnderMonitoringHours"] = "False"

                        # get and convert latest_flip_time from epoch
                        latest_flip_time_human = record.get("latest_flip_time", 0)
                        try:
                            latest_flip_time_human = float(latest_flip_time_human)
                        except:
                            latest_flip_time_human = 0
                        new_record["latest_flip_time_human"] = (
                            convert_epoch_to_datetime(latest_flip_time_human)
                        )

                        # sla_timer
                        get_sla_timer(record, sla_classes, sla_default_class)

                #
                # splk-wlk
                #

                # get record fields depending on the component
                elif self.component == "wlk":

                    # first check blocklist
                    if (
                        datagen_collection_blocklist_not_regex_dict
                        or datagen_collection_blocklist_regex_dict
                    ):
                        append_record = apply_blocklist(
                            record,
                            datagen_collection_blocklist_not_regex_dict,
                            datagen_collection_blocklist_regex_dict,
                        )

                    if append_record:

                        # get outliers
                        try:
                            isOutlier = int(record.get("isOutlier", 0))
                        except:
                            isOutlier = 0

                        try:
                            OutliersDisabled = int(record.get("OutliersDisabled", 0))
                        except:
                            OutliersDisabled = 0

                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}"'
                        )

                        # get monitoring wdays and hours ranges
                        monitoring_wdays = record.get(
                            "monitoring_wdays", "auto:all_days"
                        )
                        monitoring_hours_ranges = record.get(
                            "monitoring_hours_ranges", "auto:all_ranges"
                        )

                        # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                        (
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                        ) = get_monitoring_wdays(
                            monitoring_wdays, monitoring_hours_ranges
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", monitoring_wdays="{monitoring_wdays}", monitoring_hours_ranges="{monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                        )

                        # call get_outliers_status and define isOutlier
                        isOutlier = get_outliers_status(
                            isOutlier, OutliersDisabled, tenant_outliers_set_state
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", OutliersDisabled="{OutliersDisabled}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                        )

                        # call set_wlk_status and define object_state and anomaly_reason
                        (
                            object_state,
                            status_message,
                            status_message_json,
                            anomaly_reason,
                        ) = set_wlk_status(
                            self._metadata.searchinfo.splunkd_uri,
                            self._metadata.searchinfo.session_key,
                            self.tenant_id,
                            record,
                            isOutlier,
                            isUnderMonitoringDays,
                            isUnderMonitoringHours,
                            isUnderMonitoringMsg,
                        )
                        logging.debug(
                            f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                        )

                        # insert our main fields
                        new_record["object_state"] = object_state
                        new_record["status_message"] = " | ".join(status_message)
                        new_record["status_message_json"] = status_message_json
                        new_record["anomaly_reason"] = "|".join(anomaly_reason)

                        # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                        if isUnderMonitoringDays:
                            new_record["isUnderMonitoringDays"] = "True"
                        else:
                            new_record["isUnderMonitoringDays"] = "False"
                        if isUnderMonitoringHours:
                            new_record["isUnderMonitoringHours"] = "True"
                        else:
                            new_record["isUnderMonitoringHours"] = "False"

                        # get and convert latest_flip_time from epoch
                        latest_flip_time_human = record.get("latest_flip_time", 0)
                        try:
                            latest_flip_time_human = float(latest_flip_time_human)
                        except:
                            latest_flip_time_human = 0
                        new_record["latest_flip_time_human"] = (
                            convert_epoch_to_datetime(latest_flip_time_human)
                        )

                        # sla_timer
                        get_sla_timer(record, sla_classes, sla_default_class)

                #
                # splk-cim
                #

                # get record fields depending on the component
                elif self.component == "cim":
                    # get outliers
                    try:
                        isOutlier = int(record.get("isOutlier", 0))
                    except:
                        isOutlier = 0

                    try:
                        OutliersDisabled = int(record.get("OutliersDisabled", 0))
                    except:
                        OutliersDisabled = 0

                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}"'
                    )

                    # call get_outliers_status and define isOutlier
                    isOutlier = get_outliers_status(
                        isOutlier, OutliersDisabled, tenant_outliers_set_state
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", isOutlier="{isOutlier}", OutliersDisabled="{OutliersDisabled}", OutliersDisabled="{OutliersDisabled}", tenant_outliers_set_state="{tenant_outliers_set_state}"'
                    )

                    # get monitoring wdays and hours ranges
                    monitoring_wdays = record.get("monitoring_wdays", "auto:all_days")
                    monitoring_hours_ranges = record.get(
                        "monitoring_hours_ranges", "auto:all_ranges"
                    )

                    # call get_monitoring_wdays and define isUnderMonitoringDays, isUnderMonitoringHours, isUnderMonitoringMsg
                    (
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    ) = get_monitoring_wdays(monitoring_wdays, monitoring_hours_ranges)
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", monitoring_wdays="{monitoring_wdays}", monitoring_hours_ranges="{monitoring_hours_ranges}", isUnderMonitoringDays="{isUnderMonitoringDays}", isUnderMonitoringHours="{isUnderMonitoringHours}", isUnderMonitoringMsg="{isUnderMonitoringMsg}"'
                    )

                    # call set_cim_status and define object_state and anomaly_reason
                    (
                        object_state,
                        status_message,
                        status_message_json,
                        anomaly_reason,
                    ) = set_cim_status(
                        self._metadata.searchinfo.splunkd_uri,
                        self._metadata.searchinfo.session_key,
                        self.tenant_id,
                        record,
                        isOutlier,
                        isUnderMonitoringDays,
                        isUnderMonitoringHours,
                        isUnderMonitoringMsg,
                    )
                    logging.debug(
                        f'tenant_id="{self.tenant_id}", object_value="{object_value}", key_value="{key_value}", object_state="{object_state}", status_message="{status_message}", anomaly_reason="{anomaly_reason}"'
                    )

                    # insert our main fields
                    new_record["object_state"] = object_state
                    new_record["status_message"] = " | ".join(status_message)
                    new_record["status_message_json"] = status_message_json
                    new_record["anomaly_reason"] = "|".join(anomaly_reason)

                    # insert monitoring days and hours status (convert from 0 to False, 1 to True)
                    if isUnderMonitoringDays:
                        new_record["isUnderMonitoringDays"] = "True"
                    else:
                        new_record["isUnderMonitoringDays"] = "False"
                    if isUnderMonitoringHours:
                        new_record["isUnderMonitoringHours"] = "True"
                    else:
                        new_record["isUnderMonitoringHours"] = "False"

                    # get and convert latest_flip_time from epoch
                    latest_flip_time_human = record.get("latest_flip_time", 0)
                    try:
                        latest_flip_time_human = float(latest_flip_time_human)
                    except:
                        latest_flip_time_human = 0
                    new_record["latest_flip_time_human"] = convert_epoch_to_datetime(
                        latest_flip_time_human
                    )

                    # sla_timer
                    get_sla_timer(record, sla_classes, sla_default_class)

            #
            # End per component processing
            #

            except Exception as e:
                logging.error(
                    f'tenant_id="{self.tenant_id}", component="{self.component}", Error processing record, record="{json.dumps(record, indent=2)}", exception="{str(e)}"'
                )
                continue  # Proceed with next record

            #
            # End per component processing
            #

            if append_record:

                # add all key value pairs from the original record to new_record if not present already
                for key, value in record.items():
                    if key not in new_record:
                        new_record[key] = value

                # add new_record to processed_records
                processed_records.append(new_record)

        #
        # Render
        #

        for yield_record in self.generate_fields(processed_records):
            # logging
            logging.debug(f'yield_record="{json.dumps(yield_record, indent=2)}"')

            # yield record
            yield yield_record

        # performance counter
        logging.info(
            f'trackmedecisionmaker has terminated, tenant_id="{self.tenant_id}", component="{self.component}", upstream_records="{records_count}", processed_records="{len(processed_records)}", run_time="{round(time.time() - start, 3)}"'
        )


dispatch(TrackMeDecisionMaker, sys.argv, sys.stdin, sys.stdout, __name__)
