import os
import ast
import inspect
import sys
import logging
from logging import Logger
import coloredlogs
from pathlib import Path
import importlib.util

# Define logger
logger: Logger = logging.getLogger(__name__)
coloredlogs.install(level="DEBUG", logger=logger)

# Append path for external libraries
sys.path.append("../output/trackme/lib")

# Initialize error counter
error_count = 0


def get_func_defs(py_file):
    with open(py_file, "r") as file:
        contents = file.read()
        tree = ast.parse(contents)
        funcs = [node for node in tree.body if isinstance(node, ast.FunctionDef)]
        func_defs = {}
        for func in funcs:
            arg_names = [arg.arg for arg in func.args.args]
            default_values = [None] * (len(arg_names) - len(func.args.defaults)) + list(
                func.args.defaults
            )
            func_defs[func.name] = (arg_names, default_values, func, py_file)
    return func_defs, contents


def get_func_name(node):
    if isinstance(node.func, ast.Attribute):
        return node.func.attr
    else:
        return node.func.id


def check_arg_count(
    func_node, call_node, func_defs, script_func_defs, contents, py_file
):
    global error_count
    func_name = get_func_name(call_node)
    if func_name in func_defs:
        def_args, def_defaults, def_node, def_file = func_defs[func_name]

        # Separate handling for positional and keyword arguments in the call
        call_arg_names = [
            arg.arg for arg in call_node.args if isinstance(arg, ast.Name)
        ]
        call_kwarg_names = [kw.arg for kw in call_node.keywords]

        call_arg_count = len(call_node.args) + len(call_node.keywords)

        # Ensure all keyword arguments are actually defined in the function
        undefined_kwargs = set(call_kwarg_names) - set(def_args)
        if undefined_kwargs:
            report_error(
                func_name,
                def_args,
                def_file,
                call_node,
                contents,
                py_file,
                extra_msg=f"Undefined keyword arguments: {undefined_kwargs}",
            )
            return

        # Match arguments with defaults to calculate required and maximum arg counts
        required_arg_count = len(def_args) - len(def_defaults)
        max_arg_count = len(def_args)

        if not required_arg_count <= call_arg_count <= max_arg_count:
            report_error(func_name, def_args, def_file, call_node, contents, py_file)


def report_error(func_name, def_args, def_file, call_node, contents, py_file):
    global error_count
    error_line = contents.split("\n")[call_node.lineno - 1].strip()
    line_index = call_node.lineno
    while not error_line.strip().endswith(")"):
        line_index += 1
        error_line += " " + contents.split("\n")[line_index - 1].strip()

    logger.error(
        f"Incorrect argument count for {func_name} in file {py_file} at line {call_node.lineno}"
    )
    logger.error(
        f"Function definition (from {def_file}): def {func_name}({', '.join(def_args)}):"
    )
    logger.error(f"Function call: {error_line}")
    error_count += 1


def main():
    logger.info("Python function verification process is starting")
    lib_files = list(Path("../package/lib").rglob("trackme_libs*.py"))
    bin_files = list(Path("../package/bin").rglob("*.py"))

    all_func_defs = {}
    for lib_file in lib_files:
        logger.info(f"Loading module from file: {lib_file}")
        module_name = lib_file.stem
        module_path = str(lib_file)
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = module
        spec.loader.exec_module(module)

        func_defs, _ = get_func_defs(module_path)
        all_func_defs.update(func_defs)

        logger.info(f"Functions loaded from {module_path}: {list(func_defs.keys())}")

    for bin_file in bin_files:
        script_func_defs, contents = get_func_defs(bin_file)
        all_func_defs.update(script_func_defs)
        tree = ast.parse(contents)
        calls = [node for node in ast.walk(tree) if isinstance(node, ast.Call)]
        for call in calls:
            check_arg_count(
                None, call, all_func_defs, script_func_defs, contents, bin_file
            )

    if error_count > 0:
        logger.error(f"Found {error_count} errors.")
        sys.exit(1)
    else:
        logger.info("Python function verification completed with no errors detected.")
        sys.exit(0)


if __name__ == "__main__":
    main()
